from __future__ import annotations
import copy
import numbers
import os.path as path
import pandas as pd
from datetime import datetime
from dateutil.parser._parser import ParserError
import logging
from packaging import version
import re
from collections.abc import Iterator
import sys
try:
    from typing import Literal
except:
    from typing_extensions import Literal
import warnings

from . import data_loaders
from . import datasets
from . import __version__
from . import preproc
from . import defs
from . import exceptions
from .deprecated._decorators import deprecated, input_swap
from .datetime_parser import to_datetime

class Table:
    """
    A class that contains a DataFrame for a dataset along with meta information

    Parameters
    ----------
    details : pandas Series
        Series containing information about the dataset
    state : str
        Name of state where agencies in table are
    source_name : str
        Name of source
    agency : str
        Name of agency
    table_type : TableType enum
        Type of data contained in table
    year : int, list, MULTI
        Indicates years contained in table
    description : str
        Description of data source
    url : str
        Direct URL for accessing the table data
    source_url : str
        Main URL to view information about the data and sometimes view the data. This site contains a link to download the data.
    readme : str
        URL for data dictionary if it exists
    urls : dict
        Dictionary of url, source_url, and readme
    table : pandas of geopandasDataFrame
        Data accessed from source

    Methods
    -------
    expand(person_type="subject",
           mismatch='error')
        Expand rows that contain information on multiple subjects or officers into multiple rows
    get_csv_filename()
        Get default name of CSV file
    get_age_col(role='SUBJECT')
        Get age column (if found) after running standardize
    get_gender_col(role='SUBJECT')
        Get gender column (if found) after running standardize
    get_race_col(role='SUBJECT')
        Get race column (if found) after running standardize
    get_transform_map(orig=None, 
                      new=None, 
                      minimize=False)
        Get documentation of standardizations performed after running standardize
    merge(right, 
        std_id=False,
        keep_raw=True,
        on=None, 
        how='inner', 
        left_on=None, 
        right_on=None, 
        **kwargs)
        Merge 2 related Table objects together
    standardize(race_cats = None,
        agg_race_cat = False,
        eth_cats = None,
        gender_cats = None,
        keep_raw=True,
        known_cols = None,
        verbose = False,
        no_id = "keep",
        race_eth_combo = "merge",
        merge_date_time=True,
        empty_time = "NaT"
        Standardize column names and data values for loaded data in self.table
    to_csv(output_dir=None, filename=None)
        Convert table to CSV file
    """

    details: str= None
    state: str = None
    source_name: str = None
    agency: str = None
    table_type: defs.TableType = None
    year: int | str | list[int] = None
    description: str = None
    url: str = None
    source_url: str = None
    readme: str | None = None
    urls: dict = None

    # Data
    table: pd.DataFrame | None = None

    # Whether data is standardized
    is_std = False

    # From source
    _data_type: str = None
    _dataset_id: str | None = None
    date_field: str | None = None
    agency_field: str | None = None
    __transforms = None

    def __init__(self, 
        source: pd.DataFrame | pd.Series | dict, 
        table: pd.DataFrame | None = None, 
        year_filter: str | int | list[int] = None, 
        agency: str | None = None
        ) -> None:
        '''Construct Table object
        This is intended to be generated by the Source.load and Source.load_from_csv classes

        Parameters
        ----------
        source : pandas or geopandas Series
            Series containing information on the source
        table : pandas or geopandas 
            Name of state where agencies in table are
        '''
        if isinstance(source, pd.DataFrame):
            if len(source) == 0:
                raise LookupError("DataFrame is empty")
            elif len(source) > 1:
                raise LookupError("DataFrame has more than 1 row")

            source = source.iloc[0]

        self.details = source
        self.table = table

        self.state = source["State"]
        self.source_name = source["SourceName"]

        if agency != None:
            self.agency = agency
        else:
            self.agency = source["Agency"] if "Agency" in source else None

        try:
            self.table_type = defs.TableType(source["TableType"])  # Convert to Enum
        except:
            warnings.warn("{} is not a known table type in defs.TableType".format(source["TableType"]))
            self.table_type = source["TableType"]

        if year_filter != None:
            self.year = year_filter
        else:
            self.year = source["Year"]

        self.description = source["Description"] if "Description" in source else None
        self.url = source["URL"] if "URL" in source else None
        self._data_type = defs.DataType(source["DataType"]) if "DataType" in source else None  # Convert to Enum

        if "dataset_id" in source and not pd.isnull(source["dataset_id"]):
            self._dataset_id = source["dataset_id"]

        if "date_field" in source and not pd.isnull(source["date_field"]):
            self.date_field = source["date_field"]
        
        if "agency_field" in source and not pd.isnull(source["agency_field"]):
            self.agency_field = source["agency_field"]

        if "source_url" in source and not pd.isnull(source["source_url"]):
            self.source_url = source["source_url"]

        if "readme" in source and not pd.isnull(source["readme"]):
            self.readme = source["readme"]

        self.urls = {'source_url':self.source_url, 'readme':self.readme, 'data':self.url}


    def get_race_col(self, role:Literal['SUBJECT','OFFICER']='SUBJECT'):
        """Get name of race column

        Parameters
        ----------
        role : Literal[&#39;SUBJECT&#39;,&#39;OFFICER&#39;], optional
            Whether to return officer or subject column name, by default 'SUBJECT'

        Returns
        -------
        defs.columns
            Name of column
        """
        assert(role.upper() in ['OFFICER','SUBJECT'])
        if role.upper()=='OFFICER':
            if defs.columns.RE_GROUP_OFFICER in self.table:
                return defs.columns.RE_GROUP_OFFICER
            elif defs.columns.RE_GROUP_OFFICER_SUBJECT in self.table:
                return defs.columns.RE_GROUP_OFFICER_SUBJECT 
            else:
                return None
        else:
            if defs.columns.RE_GROUP_SUBJECT in self.table:
                return defs.columns.RE_GROUP_SUBJECT
            elif defs.columns.RE_GROUP_OFFICER_SUBJECT in self.table:
                return defs.columns.RE_GROUP_OFFICER_SUBJECT 
            else:
                return None
            

    def get_gender_col(self, role:Literal['SUBJECT','OFFICER']='SUBJECT'):
        """Get name of gender column

        Parameters
        ----------
        role : Literal[&#39;SUBJECT&#39;,&#39;OFFICER&#39;], optional
            Whether to return officer or subject column name, by default 'SUBJECT'

        Returns
        -------
        defs.columns
            Name of column
        """
        assert(role.upper() in ['OFFICER','SUBJECT'])
        if role.upper()=='OFFICER':
            if defs.columns.GENDER_OFFICER in self.table:
                return defs.columns.GENDER_OFFICER
            elif defs.columns.GENDER_OFFICER_SUBJECT in self.table:
                return defs.columns.GENDER_OFFICER_SUBJECT 
            else:
                return None
        else:
            if defs.columns.GENDER_SUBJECT in self.table:
                return defs.columns.GENDER_SUBJECT
            elif defs.columns.GENDER_OFFICER_SUBJECT in self.table:
                return defs.columns.GENDER_OFFICER_SUBJECT 
            else:
                return None

    def get_age_col(self, role:Literal['SUBJECT','OFFICER']='SUBJECT'):
        """Get name of age column

        Parameters
        ----------
        role : Literal[&#39;SUBJECT&#39;,&#39;OFFICER&#39;], optional
            Whether to return officer or subject column name, by default 'SUBJECT'

        Returns
        -------
        defs.columns
            Name of column
        """
        assert(role.upper() in ['OFFICER','SUBJECT'])
        if role.upper()=='OFFICER':
            if defs.columns.AGE_OFFICER in self.table:
                return defs.columns.AGE_OFFICER
            elif defs.columns.AGE_OFFICER_SUBJECT in self.table:
                return defs.columns.AGE_OFFICER_SUBJECT
            elif defs.columns.AGE_RANGE_OFFICER in self.table:
                return defs.columns.AGE_RANGE_OFFICER
            elif defs.columns.AGE_RANGE_OFFICER_SUBJECT in self.table:
                return defs.columns.AGE_RANGE_OFFICER_SUBJECT
            else:
                return None
        else:
            if defs.columns.AGE_SUBJECT in self.table:
                return defs.columns.AGE_SUBJECT
            elif defs.columns.AGE_OFFICER_SUBJECT in self.table:
                return defs.columns.AGE_OFFICER_SUBJECT
            elif defs.columns.AGE_RANGE_SUBJECT in self.table:
                return defs.columns.AGE_RANGE_SUBJECT
            elif defs.columns.AGE_RANGE_OFFICER_SUBJECT in self.table:
                return defs.columns.AGE_RANGE_OFFICER_SUBJECT
            else:
                return None


    def merge(self, 
              right: pd.DataFrame | pd.Series | Table, 
              std_id: bool = False,
              keep_raw: bool =True,
              on: str | pd.Index | None = None, 
              how: Literal['inner', 'outer', 'left', 'right', 'cross'] = 'inner', 
              left_on: str | pd.Index | None = None, 
              right_on: str | pd.Index | None = None, 
              **kwargs) -> Table:
        """Merge 2 Tables together or a Table with a pandas DataFrame or Series

        Parameters
        ----------
        right : pd.DataFrame | pd.Series | Table
            Table, DataFrame, or Series to merge
        std_id : bool, optional
            If True, name of incident ID column will be standardized in the merged table. by default False    
        keep_raw: bool, optional
            If std_id is True, determines whether to keep original column if incident ID column found
        on : str | pd.Index | None, optional
            Column(s) to join on. They must be found in both left and right.
            If None and left_on and right_on are None, merge will attempt to find an incident ID column to use. by default None
        how : Literal[&#39;inner&#39;, &#39;outer&#39;, &#39;left&#39;, &#39;right&#39;, &#39;cross&#39;], optional
            Type of merge to be performed. See pandas DataFrame.merge for details 
            (https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.merge.html), by default 'inner'
        left_on : str | pd.Index | None, optional
            Column(s) to join on in the left DataFrame. , by default None
        right_on : str | pd.Index | None, optional
            Column(s) to join on in the right DataFrame., by default None

        Returns
        -------
        Table
            Merged table
        """
        # Returns a new Table that is the merger of 2 tables
        if right_is_table:=isinstance(right, Table):
            df2 = right.table
        elif isinstance(right, pd.Series):
            df2 = right.to_frame()
        else:
            df2 = right

        if not self.is_std and std_id:
            raise ValueError("ID standardization is currently only possible for left tables where Table.standardize has been applied")

        use_id = not on and not left_on and not right_on
        df1 = self.table
        if use_id or std_id:
            id_col1, id_col2, df1_new, df2_new, mapping = preproc.find_id_column(self.table, df2, std_id, keep_raw)
            if use_id:
                if id_col1:
                    left_on = defs.columns.INCIDENT_ID if std_id else id_col1
                    right_on = defs.columns.INCIDENT_ID if std_id else id_col2
                else:
                    raise exceptions.AutoMergeError("Unable to automatically find ID that relates tables")
            if std_id and id_col1:
                df1 = df1_new
                df2 = df2_new
                self.__transforms.append(mapping)

        df = pd.merge(df1, df2, on=on, how=how, left_on=left_on, right_on=right_on, suffixes=(None, '_FromMerged'), **kwargs)
        new_table = copy.deepcopy(self)
        new_table.table = df
        if right_is_table and right.is_std:
            new_table._Table__transforms.extend(right.get_transform_map())
        for k,v in new_table.__dict__.items():
            if k not in ['table', '_Table__transforms']:
                setattr(new_table, k, [v, getattr(right, k) if right_is_table else None])

        return new_table



    def __repr__(self) -> str:
        skip = ["details", "table"]
        return ',\n'.join("%s: %s" % item for item in vars(self).items() if (item[0] not in skip and item[0][0] != "_"))

    def to_csv(self, 
               output_dir: str | None = None, 
               filename: str | None = None
               ) -> str:
        '''Export table to CSV file. Use default filename for data that will
        be reloaded as an openpolicedata Table object

        Parameters
        ----------
        output_dir - str
            (Optional) Output directory. Default: current directory
        filename - str
            (Optional) Filename. Default: Result of get_csv_filename()
        '''
        if filename == None:
            filename = self.get_csv_filename()
        if output_dir != None:
            filename = path.join(output_dir, filename)
        if not isinstance(self.table, pd.core.frame.DataFrame):
            raise ValueError("There is no table to save to CSV")

        self.table.to_csv(filename, index=False, errors="surrogateescape")

        return filename


    def get_csv_filename(self) -> str:
        '''Generate default filename based on table parameters

        Returns
        -------
        str
            Filename
        '''
        return get_csv_filename(self.state, self.source_name, self.agency, self.table_type, self.year)
    

    def get_transform_map(self, 
                          orig: str | None=None, 
                          new: str | None=None, 
                          minimize: bool=False):
        """Get details of standardization process if standardization has been run

        Parameters
        ----------
        minimize : bool, optional
            If True removes original value counts from output, by default False

        Returns
        -------
        Dictionary or list of dictionaries with column name and data value replacement information
        """
        
        result = copy.deepcopy(self.__transforms)
        if minimize:
            for r in result:
                del r.orig_value_counts
        
        if orig is not None or new is not None:
            for r in result:
                if (orig is None or r.orig_column_name==orig) and \
                    (new is None or r.new_column_name==new):
                    return r
            return None
        return result
    

    def expand(self, 
               person_type: Literal["subject", "officer"]="subject",
               mismatch: Literal['error','nan','splitsingle']='error'):
        """Expand demographics data into multiple row for datasets that have been standardized
         and that contain demographic and other data for more than one subject or officer per row.

        Parameters
        ----------
        person_type : Literal[&quot;subject&quot;, &quot;officer&quot;], optional
            Whether to expand subject or officer data, by default "subject"
        mismatch : Literal[&quot;error&quot;, &quot;nan&quot;, &quot;splitsingle&quot;], optional
            How to handle cases where the number of elements to be expanded is not equal across
            each column to expand. If 'error', an error will be thrown. If 'nan', columns that do not
            have enough values will be set to nan. If 'splitsingle', if the mismatched columns have a
            single value, that single value will be used for each resulting row. For example, if the
            the 2 columns of a row to be expanded are [ ['a',b'] , ['c'] ], there is a mismatch because
            the first column has 2 values and the 2nd has 1 when all columns to expand should have the
            same number. If mismatch='nan', the result will be [ 'a' , nan ] for the first column and 
            [ 'b', nan ] for the second column. If mismatch='splitsingle', the result will be 
            [ 'a' , 'c' ] for the first column and [ 'b', 'c' ] for the second column., by default "error"
        """
        
        if not self.is_std:
            raise ValueError("standardize must be run on a table to apply expand")
        
        is_subject = person_type.lower()=='subject'
        expand_cols = []
        for t in self.__transforms:
            expand = (is_subject and ("SUBJECT" in t.new_column_name or t.new_column_name==defs.columns.FATAL_SUBJECT)) or \
                        (not is_subject and "OFFICER" in t.new_column_name)
            if expand:
                if self.table[t.new_column_name].apply(lambda x: isinstance(x,dict)).any():
                    expand_cols.append(t.new_column_name)

        if len(expand_cols)>0:
            new_df = self.table.copy()
            for c in expand_cols:
                new_df[c] = new_df[c].apply(lambda x: x.values() if isinstance(x,dict) else x)

            try:
                self.table = new_df.explode(expand_cols, ignore_index=True)
            except AttributeError as e:
                # This is an issue with geopandas when exploding lists of columns
                if len(expand_cols)==1:
                    self.table = new_df.explode(expand_cols[0], ignore_index=True)
                else:
                    warnings.warn("Original table is a geopandas DataFrame, which has a known bug when expanding. "+
                                    "Converting to pandas DataFrame.")
                    self.table = pd.DataFrame(new_df).explode(expand_cols, ignore_index=True)
            except ValueError as e:
                mismatch = mismatch.lower()
                assert mismatch in ['error','nan','splitsingle']
                if mismatch!='error' and len(e.args)>0 and e.args[0]=='columns must have matching element counts':
                    dict_values = type({}.values())
                    num_vals = pd.DataFrame(0, index=new_df.index, columns=expand_cols)
                    # Try replacing length 1 arrays with their values
                    for c in expand_cols:
                        num_vals[c] = new_df[c].apply(lambda x: len(x) if isinstance(x, dict_values) else 1)
                    mismatches = num_vals.apply(lambda x: (x!=x.iloc[0]).any(), axis=1)
                    can_expand = True
                    for m in mismatches[mismatches].index:
                        max_count = num_vals.loc[m].max()
                        row = new_df.loc[m, expand_cols]
                        if mismatch=='nan':
                            fill = [pd.NA for _ in range(max_count)]
                            for k in row.index:
                                if num_vals.loc[m,k] < max_count:
                                    row[k] = fill
                        else:
                            for k in row.index:
                                if num_vals.loc[m,k] ==1:
                                    try:
                                        row[k] = [list(row[k])[0] for _ in range(max_count)]
                                    except:
                                        row[k] = [row[k] for _ in range(max_count)]
                                elif num_vals.loc[m,k]!=max_count:
                                    print("WARNING: Expansion is not possible with mismatch='splitsingle' due to 1 or more rows with "+
                                          "values of length greater than 1 and different lengths.")
                                    print(f"Example from index {m}:")
                                    print(new_df[expand_cols].loc[m])
                                    print("WARNING: Aborting expansion")
                                    can_expand = False
                                    break
                        new_df.loc[m,expand_cols] = row

                    null_locs = new_df[expand_cols].isnull().any(axis=1)
                    for m in null_locs[null_locs].index:
                        max_count = num_vals.loc[m].max()
                        row = new_df.loc[m, expand_cols]
                        if row.apply(lambda x: isinstance(x, dict_values)).any():
                            new_df.loc[m,expand_cols] = row.apply(lambda x: [x for _ in range(0,max_count)] if pd.isnull(x) else x)

                    if can_expand:
                        self.table = pd.DataFrame(new_df).explode(expand_cols, ignore_index=True)
                else:
                    raise
    

    def standardize(self, 
        race_cats: dict | str | None = None,
        agg_race_cat: bool = False,
        eth_cats: dict | None = None,
        gender_cats: dict | None = None,
        keep_raw: bool =True,
        known_cols: dict | None = None,
        verbose: bool | str = False,
        no_id: Literal["keep", "null", "error", "test"] = "keep",
        race_eth_combo: Literal[False, "merge", "concat"] = "merge",
        merge_date_time: bool =True,
        empty_time: Literal["nat", "ignore"] = "NaT"
    ):
        """Standardize column names and data values for loaded table data

        Parameters
        ----------
        race_cats : dict | str | None, optional
            Indicates data values to use for race standardization. If None, the dictionary returned by 
            opd.defs.get_race_cats() is used. If a dictionary, the keys of the dictionary must be a subset
            of the values returned by opd.defs.get_race_keys(). The corresponding values indicate, which
            value to use for each category. If race_cats is 'expand', the dictionary returned by 
            opd.defs.get_race_cats(expand=True) will be used. If race_cats is 'compact', the dictionary returned by 
            opd.defs.get_race_cats(compact=True) will be used.
        agg_race_cat : bool, optional
            If True, standardization of race will be more aggressive in converting raw values to standardized
            ones. For example, if agg_race_cat is False, standardization will not convert 'East African' to 
            the category for Black while it will if True, by default False
        eth_cats : dict | None, optional
            Indicates data values to use for ethnicity standardization. If None, the dictionary returned by 
            opd.defs.get_eth_cats() is used. If a dictionary, the keys of the dictionary must be a subset
            of the values returned by opd.defs.get_eth_keys(). The corresponding values indicate, which
            value to use for each category. If eth_cats is 'compact', the dictionary returned by 
            opd.defs.get_eth_cats(compact=True) will be used.
        gender_cats : dict | None, optional
            Indicates data values to use for gender standardization. If None, the dictionary returned by 
            opd.defs.get_gender_cats() is used. If a dictionary, the keys of the dictionary must be a subset
            of the values returned by opd.defs.get_gender_keys(). The corresponding values indicate, which
            value to use for each category. If gender_cats is 'compact', the dictionary returned by 
            opd.defs.get_gender_cats(compact=True) will be used. 
        keep_raw : bool, optional
            If False, raw columns that are standardized will be removed. If True, they will be kept and 
            renamed to indcate that they are the original raw columns, by default True
        known_cols : dict | None, optional
            Dictionary of known column mappings. If None, the dictionary defaults to 
            any known columns for this dataset ({opd.defs.columns.DATE:self.date_field, opd.defs.columns.AGENCY:self.agency_field}).
            If a dictionary, the keys of the dictionary must be available columns for standardization (defs.columns)
            and the values must be columns in the table.
        verbose : bool | str, optional
            If True, details of the standardization will be printed. If a filename, details of the standardization will
            be logged to that file., by default False
        no_id : Literal[&quot;keep&quot;, &quot;null&quot;, &quot;error&quot;], optional
            Determines how values that are not able to be standardized are handled during standardization of demographics:
            - 'keep' (default): Keep the original value
            - 'null': Replace with a null value
            - 'error': Throw an error
            , by default "keep"
        race_eth_combo : Literal[False, &quot;merge&quot;, &quot;concat&quot;], optional
            Indicates whether and how combine race and ethnicity columns. If False, race and ethnicity columns
            are not combined. If 'merge', the combined race and ethnicity column will be for Latino of all races 
            and all race categories will be for non-Latino only. Functionally, the combined race and ethnicity 
            column will be the ethnicity value if the ethnicity is Latino or unnown and the race otherwise. 
            If 'concat', race and ethnicity values will be concatenated in the combined race and ethnicity column, 
            by default "merge"
        merge_date_time : bool, optional
            If True, if standardized date and standardize time columns are identified, they will be merged into a 
            combined datetime column, by default True
        empty_time : Literal[&quot;nat&quot;, &quot;ignore&quot;], optional
            Indicates how null times are treated in the standardized datetime column (if `merge_date_time` is True). If 'NaT', the resulting 
            datetime is a null value (pd.NaT). If 'ignore', the resulting datetime will be the date value without a time., by default "NaT"

        Returns
        -------
        None
        """
        if len(self.table)==0:
            self.is_std = True
            self.__transforms = []
            return
        
        if not self.is_std:
            if known_cols is None:
                known_cols = {defs.columns.DATE:self.date_field, defs.columns.AGENCY:self.agency_field}

            race_cats = defs.get_race_cats() if race_cats is None else race_cats
            race_cats = defs.get_race_cats(expand=True) if race_cats=="expand" else race_cats
            race_cats = defs.get_race_cats(compact=True) if race_cats=="compact" else race_cats
            eth_cats = eth_cats if eth_cats is not None else defs.get_eth_cats()
            eth_cats = defs.get_eth_cats(compact=True) if eth_cats=="compact" else eth_cats
            gender_cats = gender_cats if gender_cats is not None else defs.get_gender_cats()
            gender_cats = defs.get_gender_cats(compact=True) if gender_cats=="compact" else gender_cats
            if verbose:
                logger = logging.getLogger("opd-std")
                log_level = logger.level
            if isinstance(verbose,str):
                # verbose is a filename
                fh = logging.FileHandler(verbose)
                logger.addHandler(fh)
                for handler in logger.handlers:
                    if handler.name == "main":
                        # Temporarily up level of stream handler so that only print to file
                        handler.setLevel(logging.WARNING)
            if verbose:
                # Set logger to info so log messages in preproc.standardize will be displayed
                logger.setLevel(logging.INFO)
            try:                    
                self.table, self.__transforms = preproc.standardize(self.table, self.table_type, self.year,
                    known_cols=known_cols, 
                    source_name=self.source_name,
                    state=self.state,
                    keep_raw=keep_raw,
                    agg_race_cat=agg_race_cat,
                    race_cats=race_cats,
                    eth_cats=eth_cats,
                    gender_cats=gender_cats,
                    no_id=no_id,
                    race_eth_combo=race_eth_combo,
                    merge_date_time=merge_date_time,
                    empty_time=empty_time)
            except Exception as e:
                raise e
            finally:
                if verbose:
                    logger.setLevel(log_level)
                if isinstance(verbose,str):
                    logger.removeHandler(fh)
                    for handler in logger.handlers:
                        if handler.name == "main":
                            # Revert stream handler
                            handler.setLevel(logging.NOTSET)

            self.is_std = True
        else:
            raise ValueError("Dataset has already been cleaned. Aborting cleaning.")


class Source:
    """
    Class for exploring a data source and loading its data

    ...

    Parameters
    ----------
    datasets : pandas or geopandas DataFrame
        Contains information on datasets available from the source

    Methods
    -------
    get_tables_types()
        Get types of data availble from the source
    get_years(table_type, force)
        Get years available for 1 or more datasets
    get_agencies()
        Get agencies available for 1 or more datasets
    load()
        Load data from URL
    load_from_csv()
        Load data from a previously saved CSV file
    """

    datasets: pd.DataFrame = None
    __loader = None

    def __init__(self, 
                source_name: str, 
                state: str | None =None
                ) -> None:
        '''Constructor for Source class

        Parameters
        ----------
        source_name - str
            Source name from datasets table
        state - str
            (Optional) Name of state. Only necessary if source_name is not unique.

        Returns
        -------
        Source object
        '''
        self.datasets = datasets.query(source_name=source_name, state=state)

        # Ensure that all sources are from the same state
        if len(self.datasets) == 0:
            raise ValueError(f"No Sources Found for {source_name}")
        elif self.datasets["State"].nunique() > 1:
            raise ValueError(f"There are multiple sources matching the source name {source_name}. Please specify the state of the desired location in the 2nd argument.")


    def __repr__(self) -> str:
        return str(self.datasets)


    def get_tables_types(self) -> list[str]:
        '''Get types of data availble from the source

        Returns
        -------
        list
            List containing types of data available from source
        '''
        return list(self.datasets["TableType"].unique())


    def get_years(self, 
        table_type: str | defs.TableType, 
        force: bool = False, 
        manual: bool = False,
        datasets: pd.DataFrame = None
        ) -> list[int]:
        '''Get years available for 1 or more datasets

        Parameters
        ----------
        table_type - str or TableType enum
            Only returns years for requested table type
        force - bool
            (Optional) Some data types such as CSV files require reading the whole file to filter for years. By default, an error will be thrown that indicates running load may be more efficient. For these cases, set force=True to run get_years without error.
        manual - bool
            (Optional) If True, for datasets that contain multiple years, the years will be determined by making requests to the dataset rather than using the years stored in the dataset table. The default is False, which runs faster but may not be up-to-date.
        datasets - pd.DataFrame
            (Optional) Only select from datasets in this dataframe instead of self.datasets. datasets should be a subset of the rows in self.datasets.

        Returns
        -------
        list
            List of years available for 1 or more datasets
        '''
        dfs = self.__find_datasets(table_type, datasets)

        cur_year = datetime.now().year
        all_years = list(dfs["Year"])
        years = {x for x in all_years if isinstance(x,numbers.Number) or x==defs.NA}
        for k in [k for k,x in enumerate(all_years) if x==defs.MULTI]:
            df = dfs.iloc[k]
            try:
                _check_version(df)
            except (exceptions.OPD_FutureError,exceptions.OPD_MinVersionError) as e:
                if len(years)>0:
                    # Only throw error if this is on the only dataset so far
                    continue
                else:
                    raise e
            except:
                raise
            url = df["URL"]
            date_field = df["date_field"] if pd.notnull(df["date_field"]) else None
            
            loader = self.__get_loader(df["DataType"], url, df['query'], dataset_id=df["dataset_id"], date_field=date_field)

            if not manual and pd.notnull(df["coverage_start"]) and pd.notnull(df["coverage_end"]) and \
                hasattr(df["coverage_start"], 'year') and hasattr(df["coverage_end"], 'year'):
                years.update(range(df["coverage_start"].year, df["coverage_end"].year+1))
                years_to_check = _get_years_to_check(years, cur_year, force, loader.isfile())
                if len(years_to_check)>0:
                    # Check for updates
                    new_years = loader.get_years(force=force, check=years_to_check)
                    years.update(new_years)
            else:
                new_years = loader.get_years(force=force)
                years.update(new_years)
            
        years = list(years)
        years.sort()

        return years


    def get_agencies(self, 
                     table_type: str | defs.TableType | None = None, 
                     year: str | int | None = None, 
                     partial_name: str | None = None,
                     url_contains: str | None = None,
                     id_contains: str | None = None
                     ) -> list[str]:
        '''Get agencies available for 1 or more datasets

        Parameters
        ----------
        table_type - str or TableType enum
            (Optional) If set, only returns agencies for requested table type
        year - int or the strings opd.defs.MULTI or opd.defs.NONE
            (Optional)  If set, only returns agencies for requested year
        table_type - str or TableType enum
            (Optional) If set, only returns agencies for requested table type
        partial_name - str
            (Optional)  If set, only returns agencies containing the substring
            partial_name for datasets that contain multiple agencies
        url_contains - str | None
            (Optional) If set, URL must contain this string. Can be used in combination with id_contains when multiple datasets match a set of inputs.
        id_contains - str | None
            (Optional) If set, dataset ID must contain this string. Can be used in combination with url_contains when multiple datasets match a set of inputs.

        Returns
        -------
        list
            List of agencies available for 1 or more datasets
        '''

        src = self.__find_datasets(table_type)

        if url_contains:
            src = src[src['URL'].str.contains(url_contains, regex=False)]

        if pd.notnull(id_contains):
            src = src[src['dataset_id'].str.contains(id_contains, regex=False)]

        if year != None:
            src = src[src["Year"] == year]

        if len(src) == 1:
            src = src.iloc[0]
        else:
            print(src, file=sys.stderr)
            raise ValueError("Inputs must filter for a single source")            

        # If year is opd.defs.MULTI, need to use self._agencyField to query URL
        # Otherwise return self.agency
        if src["Agency"] == defs.MULTI:
            _check_version(src)
            loader = self.__get_loader(src["DataType"], src["URL"], src['query'], dataset_id=src["dataset_id"], 
                                       date_field=src["date_field"], agency_field=src["agency_field"])
            if src["DataType"] ==defs.DataType.CSV:
                raise NotImplementedError(f"Unable to get agencies for {src['DataType']}")
            elif src['DataType'] ==defs.DataType.ArcGIS:
                raise NotImplementedError(f"Unable to get agencies for {src['DataType']}")
            elif src["DataType"] ==defs.DataType.EXCEL:
                df = loader.load(year)
                return df[src["agency_field"]].unique().tolist()
            elif src['DataType'] ==defs.DataType.SOCRATA:
                opt_filter = 'LOWER('+ src["agency_field"] + ") LIKE '%" + partial_name.lower() + "%'" if partial_name else None

                select = "DISTINCT " + src["agency_field"]
                year = None if year == defs.MULTI else year

                agency_set = loader.load(year, opt_filter=opt_filter, select=select, output_type="set")
                return list(agency_set)
            elif src['DataType'] ==defs.DataType.CKAN:
                opt_filter = 'LOWER("'+ src["agency_field"] + '")' + " LIKE '%" + partial_name.lower() + "%'" if partial_name else None

                select = 'DISTINCT "' + src["agency_field"] + '"'
                year = None if year == defs.MULTI else year

                agency_set = loader.load(year, opt_filter=opt_filter, select=select, output_type="set")
                return list(agency_set)
            else:
                raise ValueError(f"Unknown data type: {src['DataType']}")
        else:
            return [src["Agency"]]


    @input_swap([0,1], ['table_type','year'], [defs.TableType, {'values':[defs.NA, defs.MULTI], 'types':[list, int]}], opt1=None)
    def get_count(self, 
                  table_type: str | defs.TableType | None = None,
                  year: str | int | list[int] | None = None, 
                  agency: str | None = None, 
                  force: bool = False,
                  verbose: bool | str = False,
                  url_contains: str | None = None,
                  id_contains: str | None = None
                  ) -> int:
        '''Get number of records for a data request

        Parameters
        ----------
        table_type - str or TableType enum
            (Optional) If set, requested dataset will be of this type
        year (Optional) - int or length 2 list or the string opd.defs.MULTI or opd.defs.NONE
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency
        force - bool
            (Optional) For file-based data, an exception will be thrown unless force 
            is true. It may be more efficient to load the data and extract the years
            manually
        verbose - bool | str, optional
            If True, details of data loading will be logged. If a filename, details will
            be logged to that file., by default False
        url_contains - str | None
            (Optional) If set, URL must contain this string. Can be used in combination with id_contains when multiple datasets match a set of inputs.
        id_contains - str | None
            (Optional) If set, dataset ID must contain this string. Can be used in combination with url_contains when multiple datasets match a set of inputs.

        Returns
        -------
        Table
            Table object containing the requested data
        '''

        return self.__load(table_type, year, agency, True, pbar=False, return_count=True, force=force, verbose=verbose, 
                           url_contains=url_contains, id_contains=id_contains)
    
    @input_swap([0,1], ['table_type','year'], [defs.TableType, {'values':[defs.NA, defs.MULTI], 'types':[list, int]}], error=True, opt1=None)
    def load_iter(self,
                table_type: str | defs.TableType,
                year: str | int | list[int],  
                agency: str | None = None, 
                pbar: bool = False, 
                nbatch: int = 10000, 
                offset: int = 0, 
                force: bool =False,
                verbose: bool | str = False,
                url_contains: str | None = None,
                id_contains: str | None = None,
                format_date: bool = True
                ) -> Iterator[Table]:
        '''Get generator to load data from URL in batches

        Parameters
        ----------
        table_type - str or TableType enum
            Table type to load
        year - int or length 2 list or the string opd.defs.MULTI or opd.defs.NONE
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency
        pbar - bool
            (Optional) Whether to show progress bar when loading data. Default False
        nbatch - int
            (Optional) Number of records to load in each batch. Default is 10000.
        offset - int
            (Optional) Number of records to offset from first record. Default is 0 
            to return records starting from the first.
        force - bool
            (Optional) For file-based data, an exception will be thrown unless force 
            is true. It will be more efficient to read the entire dataset all at once
        verbose : bool | str, optional
            If True, details of data loading will be logged. If a filename, details will
            be logged to that file., by default False
        url_contains - str | None
            (Optional) If set, URL must contain this string. Can be used in combination with id_contains when multiple datasets match a set of inputs.
        id_contains - str | None
            (Optional) If set, dataset ID must contain this string. Can be used in combination with url_contains when multiple datasets match a set of inputs.
        format_date : bool, optional
            If True, known date columns (based on presence of date_field in datasets table or data type information provided by dataset owner) will be automatically formatted
            to be pandas datetimes (or pandas Period in rare cases), by default True

        Returns
        -------
        Table generator
            generates Table objects containing the requested data
        '''

        count = self.get_count(table_type, year, agency, force, verbose=verbose, url_contains=url_contains, id_contains=id_contains)
        for k in range(offset, count, nbatch):
            yield self.__load(table_type, year, agency, True, pbar, nrows=min(nbatch, count-k), offset=k, 
                              verbose=verbose, url_contains=url_contains, id_contains=id_contains, format_date=format_date)
    
    @deprecated("load_from_url_gen is deprecated and will be removed in a future release. Please use load_iter instead. "+
                "load_iter uses the same inputs except table_type now comes before year.")
    def load_from_url_gen(self, 
                          year: str | int | list[int], 
                          table_type: str | defs.TableType | None = None, 
                          agency: str | None = None, 
                          pbar: bool = False, 
                          nbatch: int = 10000, 
                          offset: int = 0, 
                          force: bool =False,
                          verbose: bool | str = False
                          ) -> Iterator[Table]:
        '''load_from_url_gen is deprecated. Please use load_iter instead.
        '''

        count = self.get_count(table_type, year, agency, force, verbose=verbose)
        for k in range(offset, count, nbatch):
            yield self.__load(table_type, year, agency, True, pbar, nrows=min(nbatch, count-k), offset=k, verbose=verbose)

    
    @input_swap([0,1], ['table_type','year'], [defs.TableType, {'values':[defs.NA, defs.MULTI], 'types':[list, int]}], error=True, opt1=None)
    def load(self, 
            table_type: str | defs.TableType, 
            year: str | int | list[int], 
            agency: str | None = None,
            pbar: bool = True,
            nrows: int | None = None, 
            offset: int = 0,
            sortby=None,
            verbose: bool | str = False,
            url_contains: str | None = None,
            id_contains: str | None = None,
            format_date: bool = True
            ) -> Table:
        '''Load data from URL

        Parameters
        ----------
        table_type - str or TableType enum
            Table type to load
        year - int or length 2 list or the string opd.defs.MULTI or opd.defs.NONE
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency
        pbar - bool
            (Optional) Whether to show progress bar when loading data. Default True
        nrows - int or None
            (Optional) Number of records to read. Default is None for all records.
        offset - int
            (Optional) Number of records to offset from first record. Default is 0 
            to return records starting from the first.
        verbose : bool | str, optional
            If True, details of data loading will be logged. If a filename, details will
            be logged to that file., by default False
        url_contains - str | None
            (Optional) If set, URL must contain this string. Can be used in combination with id_contains when multiple datasets match a set of inputs.
        id_contains - str | None
            (Optional) If set, dataset ID must contain this string. Can be used in combination with url_contains when multiple datasets match a set of inputs.
        format_date : bool, optional
            If True, known date columns (based on presence of date_field in datasets table or data type information provided by dataset owner) will be automatically formatted
            to be pandas datetimes (or pandas Period in rare cases), by default True

        Returns
        -------
        Table
            Table object containing the requested data
        '''

        return self.__load(table_type, year, agency, True, pbar, nrows=nrows, offset=offset, sortby=sortby, 
                           verbose=verbose, url_contains=url_contains, id_contains=id_contains, format_date=format_date)

    
    @deprecated("load_from_url is deprecated and will be removed in a future release. Please use load instead. "+
                "load uses the same inputs except table_type now comes before year.")
    def load_from_url(self, 
                      year: str | int | list[int], 
                      table_type: str | defs.TableType | None = None, 
                      agency: str | None = None,
                      pbar: bool = True,
                      nrows: int | None = None, 
                      offset: int = 0,
                      sortby=None,
                      verbose: bool | str = False
                      ) -> Table:
        '''load_from_url is deprecated and will be removed in a future release. Please use load instead.
        '''

        return self.__load(table_type, year, agency, True, pbar, nrows=nrows, offset=offset, sortby=sortby, verbose=verbose)

    def __find_datasets(self, table_type, src=None):
        if src is None:
            src = self.datasets.copy()
        elif isinstance(src, pd.Series):
            src = src.to_frame().T
        if table_type != None:
            src = src[src["TableType"].str.upper() == str(table_type).upper()]

        return src
    
    def __filter_for_source(self, table_type, year, url_contains, id_contains, errors=True):
        orig_src = self.__find_datasets(table_type)
        src = orig_src.copy()

        if isinstance(year, list) and len(year)>2:
            raise ValueError("year input must either be a single year or a list containing a start and stop year")
        
        if url_contains:
            src = src[src['URL'].str.contains(url_contains, regex=False)]

        if pd.notnull(id_contains):
            src = src[src['dataset_id'].str.contains(id_contains, regex=False)]

        matchingYears = src["Year"]==year if not isinstance(year, list) else pd.Series(False, src.index)

        if (filter_by_year:=not matchingYears.any()):
            src = src[src["Year"]==defs.MULTI]
        else:
            src = src[matchingYears]

        if isinstance(year,list):
            year_filter = []
            for y in year:
                if isinstance(y, str) and re.search(r'\d{4}-\d{2}-\d{2}', y):
                    year_filter.append(int(y[:4]))
                else:
                    year_filter.append(y)
        else:
            year_filter = year

        if len(src)>1 and year!=defs.MULTI:
            # Try to find a single multi-year dataset containing the year
            contains = pd.Series(False, index=src.index)
            for k in src.index:
                ds_years = self.get_years(table_type, datasets=src.loc[[k]])
                if isinstance(year,list):
                    if any([x>=year_filter[0] and x<=year_filter[1] for x in ds_years]):
                        contains[k] = True
                elif year in ds_years:
                    contains[k] = True
            src = src[contains]

        if len(src)>0 and isinstance(year,list) and not url_contains and pd.isnull(id_contains):
            # Ensure that year range does not also match a single year dataset
            if (orig_src['Year'].apply(lambda x: x!=defs.MULTI and x>=year_filter[0] and x<=year_filter[1])).any():
                raise ValueError(f"Year range cannot contain the year corresponding to a single year dataset.\n "
                                 f"A dataset exists for the year {year}\n "+
                                 "If the requested year range was correct, the url_contains or id_contains input can be used to specify a dataset in ambiguous cases "+
                                 "by setting url_contains to a unique substring of the desired dataset's URL or id_contains to the dataset ID. The URL(s) for the datasets matching "+
                                 f"the current inputs are {list(src['URL'])}")
            
        if isinstance(src, pd.core.frame.DataFrame):
            if len(src) == 0:
                err_msg = f"There are no sources matching {table_type=} and {year=}"
                if url_contains:
                    err_msg+=f" and {url_contains=}"
                if pd.notnull(id_contains):
                    err_msg+=f" and {id_contains=}"
                raise ValueError(err_msg)
            elif len(src) > 1:
                if errors:
                    err_msg = f"There is more than one source matching {table_type=} and {year=}"
                    if url_contains:
                        err_msg+=f" and {url_contains=}"
                    if pd.notnull(id_contains):
                        err_msg+=f" and {id_contains=}"
                    if isinstance(year, list):
                        raise ValueError(err_msg+" It is possible that the year range covers more the one dataset." +
                                        " Set the year input to not contain years for multiple datasets and/or use the url_contains or id_contains "+
                                        "input to specify a single dataset "+
                                        "by setting url_contains to a unique substring of the desired dataset's URL or id_contains to the dataset ID. "
                                        f"The URL(s) for the datasets matching the current inputs are {list(src['URL'])}")
                    elif year==defs.MULTI:
                        raise ValueError(err_msg+f" Therea are multiple multi-year datasets with year={defs.MULTI}." +
                                        " Use the url_contains or id_contains "+
                                        "input to specify a single dataset "+
                                        "by setting url_contains to a unique substring of the desired dataset's URL or id_contains to the dataset ID. "
                                        f"The URL(s) for the datasets matching the current inputs are {list(src['URL'])}")
                    else:
                        raise ValueError(err_msg+
                                        " Set the year input to a single year or a year range and/or "+
                                        "use the url_contains or id_contains input to specify a single dataset "+
                                        "by setting url_contains to a unique substring of the desired dataset's URL or id_contains to the dataset ID. "
                                        f"The URL(s) for the datasets matching the current inputs are {list(src['URL'])}")
                else:
                    # This is only for testing
                    pass
            else:
                src = src.iloc[0]

        return src, filter_by_year


    def __load(self, table_type, year, agency, load_table, pbar=True, return_count=False, force=False, 
               nrows=None, offset=0, sortby=None, verbose=False, url_contains=None, id_contains=None, format_date=True):
        # Make copy so original isn't changed
        year = year.copy() if isinstance(year, list) else year

        src, filter_by_year = self.__filter_for_source(table_type, year, url_contains, id_contains)

        # Load data from URL. For year or agency equal to opd.defs.MULTI, filtering can be done
        url = src["URL"]

        if filter_by_year:
            year_filter = year
        else:
            year_filter = None

        if not pd.isnull(src["dataset_id"]):
            dataset_id = src["dataset_id"]
        else:
            dataset_id = None

        table_year = None
        if not pd.isnull(src["date_field"]):
            date_field = src["date_field"]
            if year_filter != None:
                table_year = year_filter
        else:
            date_field = None
        
        table_agency = None
        if not pd.isnull(src["agency_field"]):
            agency_field = src["agency_field"]
            if agency != None and src['DataType'] !=defs.DataType.ArcGIS:
                table_agency = agency
        else:
            agency_field = None
        
        #It is assumed that each data loader method will return data with the proper data type so date type etc...
        if load_table:
            _check_version(src)
            loader = self.__get_loader(src['DataType'], url, src['query'], dataset_id=dataset_id, 
                                       date_field=date_field, agency_field=agency_field)

            opt_filter = None
            if agency != None and agency_field != None:
                # Double up any apostrophes for SQL query
                agency = agency.replace("'","''")
                if src['DataType']==defs.DataType.CKAN:
                    opt_filter = 'LOWER("' + agency_field + '"' + ") = '" + agency.lower() + "'"
                else:
                    opt_filter = 'LOWER(' + agency_field + ") = '" + agency.lower() + "'"

            logger = logging.getLogger("opd-load")
            if verbose:
                log_level = logger.level
            if isinstance(verbose,str):
                # verbose is a filename
                fh = logging.FileHandler(verbose)
                logger.addHandler(fh)
                for handler in logger.handlers:
                    if handler.name == "main":
                        # Temporarily up level of stream handler so that only print to file
                        handler.setLevel(logging.WARNING)
            if verbose:
                # Set logger to info so log messages in preproc.standardize will be displayed
                logger.setLevel(logging.INFO)

            logger.debug(f"Source URL: {src['source_url']}")
            logger.debug(f"Data URL: {url}")
            if dataset_id:
                logger.debug(f"]tDataset: {dataset_id}")
            logger.debug(f"Data URL: {src['DataType']}")
            if isinstance(src["readme"], str) and len(src["readme"].strip())>0:
                logger.debug(f"Data URL: {src['readme']}")
            
            try:
                if return_count:
                    return loader.get_count(year=year_filter, agency=agency, opt_filter=opt_filter, force=force)
                else:
                    table = loader.load(year=year_filter, agency=agency, opt_filter=opt_filter, nrows=nrows, pbar=pbar, offset=offset, sortby=sortby, 
                                        format_date=format_date)
                    if format_date:
                        date_field = self.__fix_date_field(table, date_field, src.name)
                        table = _check_date(table, date_field)
            except:
                raise
            finally:
                if verbose:
                    logger.setLevel(log_level)
                if isinstance(verbose,str):
                    logger.removeHandler(fh)
                    for handler in logger.handlers:
                        if handler.name == "main":
                            # Revert stream handler
                            handler.setLevel(logging.NOTSET)

        else:
            table = None

        return Table(src, table, year_filter=table_year, agency=table_agency)


    def load_from_csv(self, 
                      year: str | int | list[int],
                      output_dir: str | None = None, 
                      table_type: str | defs.TableType | None = None,
                      agency: str | None = None,
                      zip: bool =False,
                      url_contains: str | None = None,
                      id_contains: str | None = None,
                      format_date: bool = True,
                      filename: str | None = None
                      ) -> Table:
        '''Load data from previously saved CSV file
        
        Parameters
        ----------
        year - int or length 2 list or the string opd.defs.MULTI or opd.defs.NONE
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        output_dir - str
            (output_dirOptional) Directory where CSV file is stored
        table_type - str or TableType enum
            (Optional) If set, requested dataset will be of this type
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency
        zip - bool
            (Optional) Set to true if CSV is in a zip file with the same filename. Default: False
        url_contains - str | None
            (Optional) If set, URL must contain this string. Can be used in combination with id_contains when multiple datasets match a set of inputs.
        id_contains - str | None
            (Optional) If set, dataset ID must contain this string. Can be used in combination with url_contains when multiple datasets match a set of inputs.
        format_date : bool, optional
            If True, known date columns (based on presence of date_field in datasets table or data type information provided by dataset owner) will be automatically formatted
            to be pandas datetimes (or pandas Period in rare cases), by default True
        filename: str, optional
            If set, this will override the default filename based on the other inputs

        Returns
        -------
        Table
            Table object containing the requested data
        '''

        table = self.__load(table_type, year, agency, False, url_contains=url_contains, id_contains=id_contains, format_date=format_date)

        if not filename:
            filename = table.get_csv_filename()
            
        if output_dir != None:
            filename = path.join(output_dir, filename)   

        if zip:
            filename = filename.replace(".csv",'.zip')  
            
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=r"Columns \(.+\) have mixed types", category=pd.errors.DtypeWarning)
            table.table = pd.read_csv(filename, parse_dates=True, encoding_errors='surrogateescape')
        table.date_field = self.__fix_date_field(table.table, table.date_field, table.details.name)
        table.table = _check_date(table.table, table.date_field)  

        return table
    

    def find_related_tables(self,
                            table_type: str | defs.TableType | None,
                            year: str | int = None,
                            sub_type: 
                                Literal['INCIDENTS','SUBJECTS',"OFFICERS","SUBJECTS/OFFICERS",'PENALTIES','ALLEGATIONS',
                                        'NONMOTORIST','BACKGROUND','VEHICLES', None] = None,
                            exact_match: bool =False) -> tuple[str]:
        """For cases where information is split across tables (i.e. 'USE OF FORCE - INCIDENTS', 'USE OF FORCE - SUBJECTS', 'USE OF FORCE - OFFICERS'),
        find related tables for different subtypes. 'USE OF FORCE - INCIDENTS' will return 'USE OF FORCE - SUBJECTS' and 'USE OF FORCE - OFFICERS' 
        without sub_type specified and 'USE OF FORCE - SUBJECTS' if sub_type='SUBJECTS'

        Parameters
        ----------
        table_type : str | defs.TableType | None
            Type of table to find related tables for
        year - int or the strings opd.defs.MULTI or opd.defs.NONE. or None, optional
            Year of datasets to find related tables for. Optional if there is only one possible table for each table type.
        sub_type : Literal['INCIDENTS','SUBJECTS',"OFFICERS","SUBJECTS/OFFICERS",'PENALTIES','ALLEGATIONS',
                           'NONMOTORIST','BACKGROUND','VEHICLES', None], optional
            If specified, only return table containing data for the requested sub_type by default None
        exact_match : bool, optional
            If False, return tables that contain the specified sub_type (subtype="SUBJECTS" will return 'SUBJECTS' and SUBJECTS/OFFICERS tables).
            If True, return only return exact sub_type match  (subtype="SUBJECTS" will return 'SUBJECTS' table only). by default False

        Returns
        -------
        tuple[str]
            Tuple of related table types
        tuple[str]
            Tuple of year values corresponding to each table type
        """

        all_related_tables = [str(x) for x in defs.TableType if re.search(r'.+\-(?!INVOLVED)\s*', x)]
        types = set([re.sub(r'.+\-(?!INVOLVED)\s*','',x) for x in all_related_tables])
        if m:=re.search(r"^(?P<table_type>.+) - (?P<subtype>("+'|'.join(types)+r"))$", table_type, re.IGNORECASE):
            p = re.compile(rf"^{m.groupdict()['table_type']} - (?P<subtype>("+'|'.join(types)+r"))$", re.IGNORECASE)
            related = [(str(x),y) for x,y in zip(self.datasets['TableType'], self.datasets['Year']) if p.search(x) and x!=table_type]
            if sub_type:
                types = [p.search(x[0]).groupdict()['subtype'] for x in related]
                if exact_match:
                    related = [x for x,y in zip(related, types) if y==sub_type]
                else:
                    related = [x for x,y in zip(related, types) if sub_type in y]

            if year:
                related = [x for x in related if x[1]==year or x[1] in [defs.NA,defs.MULTI] or year in [defs.NA,defs.MULTI]]
            elif len(related)!=len(set([x[0] for x in related])):
                raise ValueError("Multiple possible years found for related tables. Please specify year.")
            return tuple(x for x,_ in related), tuple(x for _,x in related)
        else:
            # This table_type has no subtypes
            return (), ()


    def get_csv_filename(self, 
                         year: str | int | list[int],
                         output_dir: str | None = None, 
                         table_type: str | defs.TableType | None = None,
                         agency: str | None = None,
                         url_contains: str | None = None,
                         id_contains: str | None = None
                         ) -> str:
        '''Get auto-generated CSV filename
        
        Parameters
        ----------
        year - int or length 2 list or the string opd.defs.MULTI or opd.defs.NONE
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        output_dir - str
            (Optional) Directory where CSV file is stored
        table_type - str or TableType enum
            (Optional) If set, requested dataset will be of this type
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency
        url_contains - str | None
            (Optional) If set, URL must contain this string. Can be used in combination with id_contains when multiple datasets match a set of inputs.
        id_contains - str | None
            (Optional) If set, dataset ID must contain this string. Can be used in combination with url_contains when multiple datasets match a set of inputs.

        Returns
        -------
        str
            Auto-generated CSV filename
        '''

        table = self.__load(table_type, year, agency, False, url_contains=url_contains, id_contains=id_contains)

        filename = table.get_csv_filename()
        if output_dir != None:
            filename = path.join(output_dir, filename)             

        return filename

    def __get_loader(self, data_type, url, query, dataset_id=None, date_field=None, agency_field=None):
        if pd.isnull(dataset_id):
            dataset_id = None
        params = (data_type, url, dataset_id, date_field, agency_field)
        if self.__loader is not None and self.__loader[0]==params:
            return self.__loader[1]
        
        if dataset_id and ';' in dataset_id:
            # Multiple dataset IDs
            if data_type ==defs.DataType.CSV:
                loader = data_loaders.CombinedDataset(data_loaders.Csv, url, dataset_id, date_field=date_field, agency_field=agency_field)
            elif data_type ==defs.DataType.EXCEL:
                loader = data_loaders.CombinedDataset(data_loaders.Excel, url, dataset_id, date_field=date_field, agency_field=agency_field)
            else:
                raise ValueError(f"Not supported data type for CombinedDataset: {data_type}")
        else:
            if data_type ==defs.DataType.CSV:
                loader = data_loaders.Csv(url, date_field=date_field, agency_field=agency_field)
            elif data_type ==defs.DataType.EXCEL:
                loader = data_loaders.Excel(url, data_set=dataset_id, date_field=date_field, agency_field=agency_field) 
            elif data_type ==defs.DataType.ArcGIS:
                loader = data_loaders.Arcgis(url, date_field=date_field)
            elif data_type ==defs.DataType.SOCRATA:
                loader = data_loaders.Socrata(url, dataset_id, date_field=date_field)
            elif data_type ==defs.DataType.CARTO:
                loader = data_loaders.Carto(url, dataset_id, date_field=date_field, query=query)
            elif data_type ==defs.DataType.CKAN:
                loader = data_loaders.Ckan(url, dataset_id, date_field=date_field, query=query)
            else:
                raise ValueError(f"Unknown data type: {data_type}")

        self.__loader = (params, loader)

        return loader
    
    def __fix_date_field(self, table, date_field, loc):
        if date_field != None and table is not None and len(table)>0 and date_field not in table and \
            any([x.lower()==date_field.lower() for x in table.columns]):
            # Instances have been found where capitalization changes
            date_field = [x for x in table.columns if x.lower()==date_field.lower()][0]
            # Correct date field in tables
            self.datasets.loc[loc, "date_field"] = date_field
            datasets.datasets.loc[loc, "date_field"] = date_field

        return date_field


def _check_date(table, date_field):
    logger = logging.getLogger("opd-load")
    table = table.copy()
    if date_field != None and table is not None and len(table)>0 and date_field in table:
        dts = table[date_field]
        dts = dts[dts.notnull()]
        if len(dts) > 0:
            one_date = dts.iloc[0]  
            if isinstance(one_date, pd.Timestamp):
                if logger:
                    logger.debug(f"Converting values in column {date_field} to datetime objects")
                table[date_field] = to_datetime(table[date_field], ignore_errors=True)
            elif isinstance(one_date, str) and re.match(r'^20\d{2}(\-|/)\d{2}$',one_date):
                if logger:
                    logger.debug(f"Converting values in column {date_field} to monthly Period objects")
                table[date_field] = table[date_field].apply(pd.Period, args=('M'))
            elif (isinstance(one_date, numbers.Number) or isinstance(one_date,str)) and \
                ("year" in date_field.lower() or date_field.lower() == "yr" or \
                 (isinstance(one_date, numbers.Number) and ((dts>=1900) & (dts<2100)).all())):
                if logger:
                    logger.debug(f"Converting values in column {date_field} to annual Period objects")
                if isinstance(one_date, str) and re.search(r'\d{4}\sQ\d', one_date):
                    table[date_field] = table[date_field].str.replace(r'(\d{4}) (Q\d)', r'\1\2', regex=True)
                    table[date_field] = table[date_field].apply(pd.Period, args=('Q'))
                else:
                    table[date_field] = table[date_field].apply(pd.Period, args=('Y'))
            elif isinstance(one_date, str):
                p = re.compile(r'^Unknown string format: \d{4}-(\d{2}|__)-(\d{2}|__) present at position \d+$')
                def to_datetime_local(x):
                    try:
                        return to_datetime(x, ignore_errors=True)
                    except ParserError as e:
                        if len(e.args)>0 and p.match(e.args[0]) != None:
                            return pd.NaT
                        else:
                            raise
                    except:
                        raise

                if logger:
                    logger.debug(f"Converting values in column {date_field} to datetime objects")
                
                try:
                    # This way is much faster
                    table[date_field] = to_datetime(table[date_field], ignore_errors=True)
                except ValueError as e:
                    table[date_field] = table[date_field].apply(to_datetime_local)

                if pd.api.types.is_object_dtype(table[date_field]):
                    try:
                        # Attempt to convert
                        table = table.astype({date_field: 'datetime64[ns]'})
                    except (pd._libs.tslibs.parsing.DateParseError, ParserError) as e:
                        if 'out of range' in str(e) or \
                            re.search(r'Unknown(\sdatetime)? string format.+: \d+[-/]\d+[-/]\d+,?\s?\d+[-/]\d+[-/]\d+', str(e)):
                            pass
                        else:
                            raise
                    except UnicodeEncodeError:
                        pass
                    except ValueError as e:
                        if len(e.args)>0 and e.args[0].startswith('Cannot mix tz-aware with tz-naive values') and \
                            table[date_field].apply(lambda x: x.hour==0 and x.minute==0).all() and \
                            len(set([str(x.tzinfo) for x in table[date_field] if x.tzinfo is not None]))==1:
                            warnings.warn("Some date values have timezone and some do not. Setting timezone unware objects to have the same timezone as the rest.")
                            tz = [x.tzinfo for x in table[date_field] if x.tzinfo is not None][0]
                            is_not_aware = table[date_field].apply(lambda x: x.tzinfo is None or x.tzinfo.utcoffset(x) is None)
                            table.loc[is_not_aware, date_field] = table.loc[is_not_aware, date_field].apply(lambda x: x.replace(tzinfo=tz))
                            table[date_field] = table[date_field].convert_dtypes()
                        elif 'DateParseError' in str(type(e)) and '-__' in str(e):
                            # Ignore case where month and day are underscores (i.e. 2023-__-__)
                            pass
                        elif 'ParserError' in str(type(e)) and re.search(r'Unknown string format: \d+[-/]\d+[-/]\d+,?\s?\d+[-/]\d+[-/]\d+', str(e)):
                            # Comma separated list of dates cannot be parsed
                            pass
                        else:
                            raise
                    except TypeError as e:
                        if "Period'> is not convertible to datetime" in str(e):
                            # Mixture of datetimes and periods
                            pass
                        else:
                            raise
                    except Exception as e:
                        raise
                elif date_field.lower()=="year" and (table[date_field].dt.month==1).all() and \
                    (table[date_field].dt.day==1).all():
                    if logger:
                        logger.debug(f"Converting values in column {date_field} to annual Period objects")
                    table[date_field] = table[date_field].apply(pd.Period, args=('Y'))

                
            # Replace bad dates with NaT
            if logger:
                logger.debug(f"Replacing any values of 1900-01-01 00:00:00 in column {date_field} with null")
            table[date_field] = table[date_field].replace(datetime.strptime('1900-01-01 00:00:00', '%Y-%m-%d %H:%M:%S'), pd.NaT)

    return table


def get_csv_filename(
    state: str, 
    source_name: str, 
    agency: str, 
    table_type: str | defs.TableType, 
    year: str | int | list[int]
    ) -> str:
    '''Get default CSV filename for the given parameters. Enables reloading of data from CSV.
    
    Parameters
    ----------
    state - str
        Name of state
    source_name - str
        Name of source
    agency - str
        Name of agency
    table_type - str or TableType enum
        Type of data
    year = int or length 2 list or the string opd.defs.MULTI or opd.defs.NONE
        Year of data to load, range of years of data to load as a list [X,Y]
        to load years X to Y, or a string to indicate all of multiple year data
        (opd.defs.MULTI) or a dataset that has no year filtering ("N/A")

    Returns
    -------
    str
        Default CSV filename
    '''
    if isinstance(table_type, defs.TableType):
        table_type = table_type.value
        
    filename = f"{state}_{source_name}"
    if source_name != agency:
        filename += f"_{agency}"
    filename += f"_{table_type}"
    if isinstance(year, list):
        filename += f"_{year[0]}_{year[-1]}"
    else:
        filename += f"_{year}"

    # Clean up filename
    filename = filename.replace(",", "_").replace(" ", "_").replace("__", "_").replace("/", "_")

    filename += ".csv"

    return filename

def _check_version(df):
    min_version = df["min_version"] 
    if pd.notnull(min_version):
        src_name = df["SourceName"]
        state = df["State"]
        table_type = df["TableType"]
        year = df["Year"]
        if min_version.strip() == "-1":
            raise exceptions.OPD_FutureError(
                f"Year {year} {table_type} data for {src_name} in {state} cannot be loaded in this version. " + \
                    "It will be made available in a future release"
            )
        elif version.parse(__version__) < version.parse(min_version):
            raise exceptions.OPD_MinVersionError(
                f"Year {year} {table_type} data for {src_name} in {state} cannot be loaded in version {__version__} of openpolicedata. " + \
                    f"Update OpenPoliceData to at least version {min_version} to access this data."
            )

def _get_years_to_check(years, cur_year, force, isfile):
    max_year = max(years)
    years_to_check = []
    if  cur_year-2 <= max_year < cur_year and (force or not isfile):
        years_to_check = [x for x in range(max_year+1,cur_year+1)]

    return years_to_check



import pytest

if __name__ == "__main__":
	import sys
	sys.path.append('../openpolicedata')
from openpolicedata import data, data_loaders
from openpolicedata.defs import DataType, TableType
from openpolicedata.exceptions import OPD_DataUnavailableError, OPD_TooManyRequestsError,  \
	OPD_MultipleErrors, OPD_arcgisAuthInfoError, OPD_SocrataHTTPError, OPD_FutureError, OPD_MinVersionError
import openpolicedata as opd
from datetime import datetime
import pandas as pd
from time import sleep
import warnings
import os

import pathlib
import sys
sys.path.append(pathlib.Path(__file__).parent.resolve())
from test_utils import check_for_dataset

sleep_time = 0.1

# Set Arcgis data loader to validate queries with arcgis package if installed
data_loaders._verify_arcgis = True

log_filename = f"pytest_url_errors_{datetime.now().strftime('%Y%m%d_%H')}.txt"
log_folder = os.path.join(".","data/test_logs")

outages_file = os.path.join("..","opd-data","outages.csv")
# if has_outages:=os.path.exists(outages_file):
has_outages=os.path.exists(outages_file)
if has_outages:
	outages = pd.read_csv(outages_file)
else:
	try:
		outages = pd.read_csv('https://raw.githubusercontent.com/openpolicedata/opd-data/main/outages.csv')
		has_outages = True
	except:
		pass

warn_errors = (OPD_DataUnavailableError, OPD_SocrataHTTPError, OPD_FutureError, OPD_MinVersionError)

def user_request_skip(datasets, i, skip, start_idx, source):
	# Skip sources that the user requested to skip
	if skip and datasets.iloc[i]["SourceName"] in skip:
		return True
	# User requested to start at start_idx
	if i<start_idx:
		return True
	if source != None and datasets.iloc[i]["SourceName"] != source:
		return True
	
	return False

def check_table_type_warning(all_datasets):
	sources = all_datasets.copy().iloc[0]
	sources["TableType"] = "TEST"
	with pytest.warns(UserWarning):
		data.Table(sources)

@pytest.mark.parametrize('source, table, year', [('Phoenix', "OFFICER-INVOLVED SHOOTINGS", 2022), 
				('Orlando', "OFFICER-INVOLVED SHOOTINGS", 2022), ('Indianapolis', "OFFICER-INVOLVED SHOOTINGS", 2022),
				('Philadelphia', "OFFICER-INVOLVED SHOOTINGS", 2018)])
def test_format_date_false(all_datasets, source, table, year):
	if check_for_dataset(source, table):
		src = opd.Source(source)
		table = src.load(table, year, format_date=False, nrows=1)
		# Confirm date has not been formatted
		assert isinstance(table.table[table.date_field].iloc[0],str)
		

@pytest.mark.parametrize('source, table, year', [('Denver', "OFFICER-INVOLVED SHOOTINGS", 2022), 
				('Sparks', "OFFICER-INVOLVED SHOOTINGS", 2022),  ('Louisville', "TRAFFIC STOPS", ['2018-12-29', '2019-01-01'])])
def test_format_date_false_not_allowed(all_datasets, source, table, year):
	if check_for_dataset(source, table):
		src = opd.Source(source)
		with pytest.raises(ValueError, match='Dates cannot be filtered'):
			src.load(table, year, format_date=False, nrows=1)


def test_check_version(datasets):
	ds = datasets.iloc[0].copy()
	# Set min_version to create error
	ds["min_version"] = "-1"
	with pytest.raises(OPD_FutureError):
		data._check_version(ds)

	ds["min_version"] = "100000.0"
	with pytest.raises(OPD_MinVersionError):
		data._check_version(ds)

	# These should pass
	ds["min_version"] = "0.0"
	data._check_version(ds)
	ds["min_version"] = pd.NA
	data._check_version(ds)

def test_single_year_filter():
	if check_for_dataset('Phoenix', opd.defs.TableType.CALLS_FOR_SERVICE):
		src = data.Source('Phoenix')
		year = 2017
		dataset, filter_by_year = src._Source__filter_for_source(opd.defs.TableType.CALLS_FOR_SERVICE, year, None, None)
		assert not filter_by_year
		assert dataset['Year']==year
		assert dataset['TableType']==opd.defs.TableType.CALLS_FOR_SERVICE

def test_single_year_range_filter():
	if check_for_dataset('Phoenix', opd.defs.TableType.CALLS_FOR_SERVICE):
		src = data.Source('Phoenix')
		with pytest.raises(ValueError, match="There are no sources matching"):
			src._Source__filter_for_source(opd.defs.TableType.CALLS_FOR_SERVICE, [2017,2018], None, None)

@pytest.mark.parametrize('year, exp_filt_by_year', [(2017,True), (opd.defs.MULTI,False)])
def test_multi_year_filter(year,exp_filt_by_year):
	if check_for_dataset('Norristown', opd.defs.TableType.USE_OF_FORCE):
		src = data.Source('Norristown')
		dataset, filter_by_year = src._Source__filter_for_source(opd.defs.TableType.USE_OF_FORCE, year, None, None)
		assert filter_by_year == exp_filt_by_year
		assert dataset['Year']==opd.defs.MULTI
		assert dataset['TableType']==opd.defs.TableType.USE_OF_FORCE

def test_multi_year_range_filter_includes_single_year():
	if check_for_dataset('Norristown', opd.defs.TableType.USE_OF_FORCE):
		src = data.Source('Norristown')
		year = [2016,2018]
		with pytest.raises(ValueError, match="Year range cannot contain the year corresponding to a single year"):
			src._Source__filter_for_source(opd.defs.TableType.USE_OF_FORCE, year, None, None)

def test_bad_year_range():
	if check_for_dataset('Norristown', opd.defs.TableType.USE_OF_FORCE):
		src = data.Source('Norristown')
		year = [2016,2018, 2017]
		with pytest.raises(ValueError, match="year input must either be a single year or a list containing a start and stop year"):
			src._Source__filter_for_source(opd.defs.TableType.USE_OF_FORCE, year, None, None)

@pytest.mark.parametrize('year, url_contains', [(2019, "APDUseOfForce/FeatureServer/0"), 
												([2018,2019], "APDUseOfForce/FeatureServer/0"), (2021, "APD_UseOfForce2021/FeatureServer/0")])
def test_multiple_multiple_year_filter(all_datasets, year, url_contains):
	if check_for_dataset('Asheville', opd.defs.TableType.USE_OF_FORCE):
		src = data.Source('Asheville')
		dataset, filter_by_year = src._Source__filter_for_source(opd.defs.TableType.USE_OF_FORCE, year, None, None)
		assert filter_by_year
		assert dataset['Year']==opd.defs.MULTI
		assert dataset['TableType']==opd.defs.TableType.USE_OF_FORCE
		assert url_contains in dataset['URL']

@pytest.mark.parametrize('year, url_contains, url_contains_test', [(2020, None, 'LMPD_STOPS_DATA_(2)'), 
												(2021, None, 'LMPD_STOP_DATA_2021'),
												(2021, 'LMPD_STOPS_DATA_(2)', 'LMPD_STOPS_DATA_(2)')])
def test_overlapping_multi_and_single(all_datasets, year, url_contains, url_contains_test):
	if check_for_dataset('Louisville', opd.defs.TableType.TRAFFIC):
		src = data.Source('Louisville')
		dataset, filter_by_year = src._Source__filter_for_source(opd.defs.TableType.TRAFFIC, year, url_contains, None)
		assert url_contains_test in dataset['URL']

@pytest.mark.parametrize('year, id_contains, id_contains_test', [(2020, None, 'ex94-c5ad'), 
												(2021, None, 'izhu-764k'),
												([2020,2021], 'izhu-764k', 'izhu-764k')])
def test_overlapping_multi_and_single_with_id(all_datasets, year, id_contains, id_contains_test):
	if check_for_dataset('Mesa', opd.defs.TableType.CALLS_FOR_SERVICE):
		src = data.Source('Mesa')
		dataset, filter_by_year = src._Source__filter_for_source(opd.defs.TableType.CALLS_FOR_SERVICE, year, None, id_contains)
		assert id_contains_test == dataset['dataset_id']

@pytest.mark.parametrize('year', [2020, [2020,2021], opd.defs.MULTI])
def test_multiple_bad_multiple_year_filter(all_datasets, year):
	if check_for_dataset('Asheville', opd.defs.TableType.USE_OF_FORCE):
		src = data.Source('Asheville')
		with pytest.raises(ValueError, match="There is more than one source matching "):
			src._Source__filter_for_source(opd.defs.TableType.USE_OF_FORCE, year, None, None)

@pytest.mark.parametrize('year, exp_filt_by_year', [([2021,2022], True), (opd.defs.MULTI, False)])
def test_url_contains(all_datasets, year, exp_filt_by_year):
	if check_for_dataset('Asheville', opd.defs.TableType.USE_OF_FORCE):
		src = data.Source('Asheville')
		url_contains = "APDUseOfForce/FeatureServer/0"
		dataset, filter_by_year = src._Source__filter_for_source(opd.defs.TableType.USE_OF_FORCE, year, url_contains, None)
		assert filter_by_year == exp_filt_by_year
		assert dataset['Year']==opd.defs.MULTI
		assert dataset['TableType']==opd.defs.TableType.USE_OF_FORCE
		assert url_contains in dataset['URL']


def test_offsets_and_nrows():
	if check_for_dataset('Philadelphia', opd.defs.TableType.SHOOTINGS):
		src = data.Source("Philadelphia")
		df = src.load(year=2019, table_type="Officer-Involved Shootings").table
		offset = 1
		nrows = len(df)-2
		df_offset = src.load(year=2019, table_type="Officer-Involved Shootings", offset=offset, nrows=nrows).table
		assert df_offset.equals(df.iloc[offset:offset+nrows].reset_index(drop=True))

def check_excel_sheets(datasets, source, start_idx, skip):
	for i in range(len(datasets)):
		if user_request_skip(datasets, i, skip, start_idx, source):
			continue

		if datasets.iloc[i]["DataType"]!=DataType.EXCEL:
			continue

		srcName = datasets.iloc[i]["SourceName"]
		state = datasets.iloc[i]["State"]
		src = data.Source(srcName, state=state)

		table_print = datasets.iloc[i]["TableType"]
		now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
		print(f"{now} Testing {i+1} of {len(datasets)}: {srcName} {table_print} table")

		excel = src._Source__get_loader(datasets.iloc[i]["DataType"], datasets.iloc[i]["URL"], datasets.iloc[i]["query"],
				dataset_id=datasets.iloc[i]["dataset_id"])
		sheets, has_year_sheets = excel._Excel__get_sheets()

		if has_year_sheets:
			# Ensure that load works
			src.load(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], pbar=False)
		else:
			excel._Excel__check_sheet(sheets)


def test_source_download_limitable(datasets, source, start_idx, skip, loghtml, query={}):
	num_stanford = 0
	max_num_stanford = 1  # This data is standardized. Probably no need to test more than 1
	caught_exceptions = []
	caught_exceptions_warn = []
		
	for i in range(len(datasets)):
		if user_request_skip(datasets, i, skip, start_idx, source):
			continue

		has_date_field = not pd.isnull(datasets.iloc[i]["date_field"])
		if can_be_limited(datasets.iloc[i]["DataType"], datasets.iloc[i]["URL"]) or has_date_field:
			if is_stanford(datasets.iloc[i]["URL"]):
				num_stanford += 1
				if num_stanford > max_num_stanford:
					continue
			match = True
			for k,v in query.items():
				if datasets.iloc[i][k]!=v:
					match = False
					break
			if not match:
				continue

			srcName = datasets.iloc[i]["SourceName"]
			state = datasets.iloc[i]["State"]
			src = data.Source(srcName, state=state)

			table_print = datasets.iloc[i]["TableType"]
			now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
			print(f"{now} Testing {i} of {len(datasets)-1}: {srcName} {table_print} table")

			nrows = 20 if datasets.iloc[i]['DataType']!='Excel' else None

			# Handle cases where URL is required to disambiguate requested dataset
			ds_filter, _ = src._Source__filter_for_source(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], None, None, errors=False)
			url_contains = datasets.iloc[i]['URL'] if isinstance(ds_filter,pd.DataFrame) and len(ds_filter)>1 else None
			id_contains = datasets.iloc[i]['dataset_id'] if isinstance(ds_filter,pd.DataFrame) and len(ds_filter)>1 else None

			try:
				table = src.load(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], pbar=False, nrows=nrows, 
					 url_contains=url_contains, id_contains=id_contains)
			except warn_errors as e:
				e.prepend(f"Iteration {i}", srcName, datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"])
				caught_exceptions_warn.append(e)
				continue
			except (OPD_TooManyRequestsError, OPD_arcgisAuthInfoError) as e:
				# Catch exceptions related to URLs not functioning
				e.prepend(f"Iteration {i}", srcName, datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"])
				caught_exceptions.append(e)
				continue
			except:
				raise

			if len(table.table)==0 and has_outages and \
				(outages[["State","SourceName","Agency","TableType","Year"]] == datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]]).all(axis=1).any():
				caught_exceptions_warn.append(f'Outage continues for {str(datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]])}')
				continue
			else:
				assert len(table.table)>0

			if isinstance(datasets.iloc[i]['dataset_id'], str) and ';' in datasets.iloc[i]['dataset_id']:
				# A year-long dataset is generated from multiple datasets. Check that whole year is covered
				# Parse dataset ID to get number of datasets IDs
				sheets = None
				dataset_list = datasets.iloc[i]['dataset_id']
				if '|' in dataset_list:  # dataset names are separated from relative URLs by |
					dataset_list = dataset_list.split('|')
					assert len(dataset_list)==2
					sheets = dataset_list[0].split(';')   # Different sheet names for each dataset are separated by ;. If multiple sheets for a given dataset, separate by &
					dataset_list = dataset_list[1]
				dataset_list = dataset_list.split(';')  
				
				table.standardize()

				assert opd.defs.columns.DATE in table.table
				months = table.table[opd.defs.columns.DATE].dt.month.unique()
				if len(months)!=12:
					missing_months = [x for x in range(1,13) if x not in months]
					if srcName=='Wallkill' and datasets.iloc[i]["Year"]==2016:
						# Only has last 3 quarters of data
						missing_months = [x for x in missing_months if x not in [1,2,3]]
					elif srcName=='Albemarle County' and datasets.iloc[i]["Year"]==2021:
						# August and Sept data does not have date field
						missing_months = [x for x in missing_months if x not in [8,9]]
					for m in missing_months:
						url = datasets.iloc[i]["URL"] + '/' + dataset_list[m-1].strip()
						cur_sheet = sheets[min(m-1, len(sheets)-1)] if sheets else None
						loader = data_loaders.Excel(url, cur_sheet, datasets.iloc[i]["date_field"])
						df_check = loader.load()
						assert df_check['DATE'].apply(lambda x: (int(x[:2]) if isinstance(x,str) else x.month) !=m).all()

			if not pd.isnull(datasets.iloc[i]["date_field"]):
				if datasets.iloc[i]["date_field"] not in table.table:
					table = src.load(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], pbar=False, nrows=2000)
				assert datasets.iloc[i]["date_field"] in table.table
				#assuming a Pandas string dtype('O').name = object is okay too
				assert (table.table[datasets.iloc[i]["date_field"]].dtype.name in ['datetime64[ns]', 'datetime64[ns, UTC]', 
																	'datetime64[ms]', 'period[Y-DEC]','period[Q-DEC]',
																	'period[M]']) or \
						table.table[datasets.iloc[i]["date_field"]].apply(lambda x: type(x) in [pd.Timestamp,pd.Period]).mean()>0.9
				dts = table.table[datasets.iloc[i]["date_field"]]
				dts = dts[dts.notnull()]
				# New Orleans complaints dataset has many empty dates
				# "Seattle and Minneapolis starts with bad date data"
				if len(dts)>0 or srcName not in ["Seattle","New Orleans",'Minneapolis','St. Paul'] or \
					datasets.iloc[i]["TableType"] not in [TableType.COMPLAINTS, TableType.INCIDENTS]:
					assert len(dts) > 0   # If not, either all dates are bad or number of rows requested needs increased
					assert dts.iloc[0].year <= datetime.now().year if isinstance(dts.iloc[0], (pd.Timestamp,pd.Period)) else \
						dts.iloc[1].year <= datetime.now().year
			if not pd.isnull(datasets.iloc[i]["agency_field"]):
				assert datasets.iloc[i]["agency_field"] in table.table

			# Adding a pause here to prevent issues with requesting from site too frequently
			sleep(sleep_time)

	if loghtml:
		log_errors_to_file(caught_exceptions, caught_exceptions_warn)
	else:
		if len(caught_exceptions)==1:
			raise caught_exceptions[0]
		elif len(caught_exceptions)>0:
			msg = f"{len(caught_exceptions)} URL errors encountered:\n"
			for e in caught_exceptions:
				msg += "\t" + e.args[0] + "\n"
			raise OPD_MultipleErrors(msg)

		for e in caught_exceptions_warn:
			warnings.warn(str(e))

@pytest.mark.parametrize("loader, source, table_type, agency, years", [
	(data_loaders.Socrata, 'Richmond', 'CALLS FOR SERVICE', None, [2021, [2020, 2022]]),
	(data_loaders.Ckan, 'Virginia', 'STOPS', "Arlington County Police Department", [2021, [2020, 2022]]),
	(data_loaders.Arcgis, "Charlotte-Mecklenburg", 'EMPLOYEE', None, []),
	(data_loaders.Csv, 'Denver', "OFFICER-INVOLVED SHOOTINGS", None, []),
	(data_loaders.Excel, 'Rutland', "USE OF FORCE", None, []),
	(data_loaders.Carto, "Philadelphia", 'STOPS', None, [2021, [2020, 2022]])
	])
def test_get_count(loader, source, table_type, agency, years):
	if check_for_dataset(source, table_type):
		print(f"Testing {loader} source")
		src = opd.Source(source)
		i = src.datasets["TableType"] == table_type
		i = i[i].index[0]
		if pd.notnull(src.datasets.loc[i]["dataset_id"]):
			loader = loader(src.datasets.loc[i]["URL"], src.datasets.loc[i]["dataset_id"], date_field=src.datasets.loc[i]["date_field"])
		else:
			loader = loader(src.datasets.loc[i]["URL"], date_field=src.datasets.loc[i]["date_field"])
		if len(years)==0:
			assert loader.get_count(force=True) == src.get_count(year=src.datasets.loc[i]['Year'], table_type=table_type, force=True)
		else: 
			for year in years:
				assert loader.get_count(year=year, force=True) == src.get_count(year=year, table_type=table_type, force=True)

		if agency:
			opt_filter = '"' + src.datasets.loc[i]["agency_field"] + '"' + " = '" + agency + "'"
			year = years[0]
			assert src.get_count(year=year, agency=agency, force=True) == loader.get_count(year=year, opt_filter=opt_filter, force=True)


def test_get_years_to_check():
	assert data._get_years_to_check([2020], cur_year=2023, force=True, isfile=False) == []
	assert data._get_years_to_check([2022], cur_year=2023, force=False, isfile=True) == []
	assert data._get_years_to_check([2022, 2020], cur_year=2023, force=False, isfile=False) == [2023]
	assert data._get_years_to_check([2020, 2021], cur_year=2023, force=True, isfile=True) == [2022, 2023]
	assert data._get_years_to_check([2020, 2021], cur_year=2023, force=True, isfile=False) == [2022, 2023]


@pytest.mark.parametrize('source,year,table_type,nrows,agency', [
	("Virginia",2020,"STOPS", 2000, "Fairfax County Police Department"), # CKAN
	("Philadelphia", 2021, "STOPS", 1000, None),  # Carto
	("Richmond","MULTIPLE","OFFICER-INVOLVED SHOOTINGS", 5, None), # Socrata
	("Fairfax County",2016,"ARRESTS", 1000, None),  # ArcGIS
	("Norristown", 2016, "USE OF FORCE",100, None), # Excel
	("Denver", "MULTIPLE", "OFFICER-INVOLVED SHOOTINGS", 50, None) # CSV
])
def test_load_gen(source, year, table_type, nrows, agency):
	if check_for_dataset(source, table_type):
		src = data.Source(source)
		max_iter = 10
		df = src.load(table_type, year, agency=agency, nrows=max_iter*nrows).table
		with warnings.catch_warnings():
			warnings.filterwarnings("ignore",category=RuntimeWarning)
			df = df.convert_dtypes()

		offset = 0
		k = 0
		for t in src.load_iter(table_type, year, nbatch=nrows, force=True, agency=agency):
			df_cur = df.iloc[offset:offset+len(t.table)].reset_index(drop=True)
			df2 = t.table.copy()
			if set(df.columns)!=set(df2.columns):
				# Expecting that a column was not returned because it is empty in the subset requested in this iteration
				assert len(df2.columns)<len(df.columns)
				missing_columns = list(set(df.columns).difference(set(df2.columns)))
				for col in missing_columns:
					assert df_cur[col].isnull().all()
					df_cur.drop(columns=col, inplace=True)
			# Assure columns are in proper order
			assert set(df_cur.columns)==set(df2.columns)
			df2 = df2[df_cur.columns]
			# Ensure that dtypes match
			df2 = df2.astype(df_cur.dtypes.to_dict())
			assert df2.equals(df_cur)
			offset+=len(t.table)
			k+=1
			if k>=max_iter:
				break


def can_be_limited(data_type, url):
	if (data_type == DataType.CSV and ".zip" in url):
		return False
	elif data_type in [DataType.ArcGIS, DataType.SOCRATA, DataType.CSV, DataType.EXCEL, DataType.CARTO, DataType.CKAN]:
		return True
	else:
		raise ValueError("Unknown table type")
	

def is_stanford(url):
	return "stanford.edu" in url

def log_errors_to_file(*args):
	if not os.path.exists(log_folder):
		os.mkdir(log_folder)

	filename = os.path.join(log_folder, log_filename)

	if os.path.exists(filename):
		perm = "r+"
	else:
		perm = "w"

	with open(filename, perm) as f:
		for x in args:
			for e in x:
				new_line = ', '.join([str(x) for x in e.args])
				skip = False
				if perm == "r+":
					for line in f:
						if new_line in line or line in new_line:
							skip = True
							break

				if not skip:
					f.write(new_line)
					f.write("\n")

if __name__ == "__main__":
	from test_utils import get_datasets
	# For testing
	use_changed_rows = False
	csvfile = None
	# csvfile = os.path.join(r"..",'opd-data','opd_source_table.csv')
	start_idx = 0
	skip = None
	# skip = "Sacramento"
	source = None
	# source = "Bremerton" #"Washington D.C." #"Wallkill"
	query = {}
	# query = {'DataType':'Excel'}

	datasets = get_datasets(csvfile, use_changed_rows)

	# check_excel_sheets(csvfile, source, last, skip, None) 
	# test_get_years_to_check(csvfile, source, last, skip, None) 
	# check_table_type_warning(csvfile, source, last, skip, None) 
	# test_offsets_and_nrows(csvfile, source, last, skip, None) 
	# test_check_version(csvfile, None, last, skip, None) #
	test_source_download_limitable(datasets, source, start_idx, skip, False, query) 
	
	# test_get_count(csvfile, None, last, skip, None)
	# test_load_gen(csvfile, source, last, skip, None) 
	



from ..datasets import query as _query
from ._decorators import deprecated

# For back compatibility, preferred usage is datasets.query
@deprecated("opd.datasets_query is deprecated and will be removed in a future version. "
        "Please use opd.datasets.query instead.")
def datasets_query(source_name=None, state=None, agency=None, table_type=None):
    return _query(source_name=source_name, state=state, agency=agency, table_type=table_type)


from packaging import version
import pandas as pd
import warnings

from ..exceptions import CompatSourceTableLoadError
from .. import __version__


def check_compat_source_table(column_types=None, df_compat=None, cur_ver=__version__):
    compat_versions_file = 'https://github.com/openpolicedata/opd-data/raw/main/compatibility/compat_versions.csv'
    idx = compat_versions_file.rfind('/')
    df = None
    success = False
    try:
        cur_ver = version.parse(cur_ver)
        if df_compat is None:
            df_compat = pd.read_csv(compat_versions_file, dtype=str)
        df_compat['version'] = df_compat['version'].apply(version.parse)
        df_compat = df_compat.sort_values(by='version')
        df_compat['required'] = df_compat['required'].apply(lambda x: x=='True')
        avail_ver = df_compat['version'].apply(lambda x: cur_ver<=x)
    except:
        return success, df
        
    if avail_ver.any():
        for k in avail_ver[avail_ver].index:
            try:
                file = compat_versions_file[:idx+1]+df_compat['csv_name'].loc[k]
                df = pd.read_csv(file, dtype=column_types)
                success = True
                warnings.warn("This version of OpenPoliceData requires usage of the deprecated source table located at "+
                        f"{file}. OpenPoliceData will still operate with this source table. "+
                        "However, the latest datasets added to OpenPoliceData will not be available. "
                        "Updating OpenPoliceData is recommended: python -m pip install openpolicedata --upgrade.",
                        DeprecationWarning)
                break
            except:
                if df_compat['required'].loc[k]:
                    raise CompatSourceTableLoadError("This version of OpenPoliceData requires usage of the deprecated source table located at "+
                                                    f"{file}. This file cannot be loaded. This may be due to a poor internet connection. "+
                                                    "If not, updating OpenPoliceData is recommended: python -m pip install openpolicedata --upgrade.")
                
    return success, df


import datetime
import dateutil
import math
import numpy as np
import pandas as pd
from pandas .api.types import is_numeric_dtype
import datetime as dt
import re
import warnings
import numbers

from .utils import is_str_number

def parse_date_to_datetime(date_col):
    if len(date_col.shape)==2:
        if date_col.shape[1] > 1:
            dts = date_col.iloc[:,0][date_col.iloc[:,0].notnull()]
            if hasattr(dts.iloc[0], "year"):
                if isinstance(dts.dtype, pd.api.types.PeriodDtype):
                    un_vals = to_datetime([x.to_timestamp() for x in dts.unique()])
                else:
                    un_vals = to_datetime(dts.unique())
                if (un_vals.month != 1).any() or (un_vals.day != 1).any() or (un_vals.hour != 0).any() or \
                    (un_vals.minute != 0).any() or (un_vals.second != 0).any():
                    raise ValueError("Expected year data to not contain any month, day, or time info")

                # Making a copy to avoid warning
                d = date_col.copy()
                d[d.columns[0]] = date_col.iloc[:,0].dt.year

                def month_name_to_num(x):
                    month_list = ["jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"]
                    if isinstance(x,str) and not x.isdigit():
                        month_num = [k+1 for k,y in enumerate(month_list) if x.lower().startswith(y)]
                        return month_num[0]
                    else:
                        return int(x) if pd.notnull(x) else x

                d[d.columns[1]] = date_col.iloc[:,1].apply(month_name_to_num)

                if d.shape[1]==2:
                    return d.apply(lambda x: pd.Period(to_datetime(f"{int(x.iloc[1])}/1/{int(x.iloc[0])}"), 'M') 
                                   if pd.notnull(x.iloc[1])
                                   else pd.Period(to_datetime(f"{int(x.iloc[0])}"), 'Y'), axis=1)
                else:
                    return to_datetime(d)
        else:
            date_col = date_col.iloc[:,0]

    ind_is_num = date_col.notnull()
    if ind_is_num.any():
        if not hasattr(date_col[ind_is_num].iloc[0], "year") or (date_col[ind_is_num].apply(type)==str).any():
            is_num = is_numeric_dtype(date_col)
            if not is_num:
                # Try to convert to all numbers
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=RuntimeWarning, message="invalid value encountered in cast")
                    new_col = date_col.convert_dtypes()
                if new_col.dtype in ["object", "string", "string[python]"] and \
                    new_col.apply(lambda x: pd.notnull(x) and (isinstance(x, numbers.Number) or is_str_number(x))).sum() > 0 and \
                    new_col.apply(lambda x: pd.isnull(x) or isinstance(x,(pd.Timestamp,int, dt.datetime)) or \
                                  isinstance(x, numbers.Number) or is_str_number(x) or \
                                  (isinstance(x,str) and x.strip()=="")).sum()>=len(new_col)-1:
                    # Almost all numeric, null, or timestamp values and at least one numeric value
                    def clean_dates(x):
                        if pd.notnull(x) and (isinstance(x, numbers.Number) or is_str_number(x)):
                            return int(float(x))
                        elif isinstance(x,str):
                            return np.nan
                        else:
                            return x
                        
                    date_col = new_col.apply(clean_dates)

                    ind_is_num = date_col.notnull() & date_col.apply(lambda x: not isinstance(x, (pd.Timestamp, pd.Period, dt.datetime)))
                    is_num = True

            if is_num:
                # Date as number like MMDDYYYY. Need to determine order
                # Assuming year is either first or last
                dts = date_col[ind_is_num]
                if (dts < 0).any():
                    raise ValueError("Date values cannot be negative")

                year_last = dts % 10000
                year_first = np.floor(dts / 10000)
                year_last_2digit = dts % 100
                this_year = dt.datetime.now().year

                is_valid_last = (year_last <= this_year).all() and (year_last > 1300).all()
                is_valid_first = (year_first <= this_year).all() and (year_first > 1300).all()
                is_valid_last_2digit = (year_last_2digit <= this_year-2000).all() and (year_last_2digit >= 0).all()

                any_valid = True
                if is_valid_first:
                    year = dts.apply(lambda x : np.floor(x / 10000) if not pd.isnull(x) else x)
                    month_day = dts.apply(lambda x : x % 10000 if not pd.isnull(x) else x)
                elif is_valid_last:
                    year = dts.apply(lambda x : x % 10000 if not pd.isnull(x) else x)
                    month_day = dts.apply(lambda x : np.floor(x / 10000) if not pd.isnull(x) else x)
                elif is_valid_last_2digit:
                    year = 2000+dts.apply(lambda x : x % 100 if not pd.isnull(x) else x)
                    month_day = dts.apply(lambda x : np.floor(x / 100) if not pd.isnull(x) else x)
                else:
                    any_valid = False

                if any_valid:
                    # Determine if month is first or last in month_day
                    first_val = month_day.apply(lambda x: np.floor(x/100) if pd.notnull(x) and isinstance(x,numbers.Number) else x)
                    last_val = month_day % 100

                    is_valid_month_first = first_val.max() < 13 and last_val.max() < 32
                    is_valid_month_last = last_val.max() < 13 and first_val.max() < 32
                    if is_valid_month_first:
                        # If also is_valid_month_last, assuming month is first because this is more common in the U.S. 
                        month = first_val
                        day = last_val
                    elif is_valid_month_last:
                        month = last_val
                        day = first_val
                    else:
                        any_valid = False

                    if any_valid:
                        # Convert data type to object to avoid changing dtype warning and then convert to TimeStamp
                        date_col = date_col.astype('O')
                        date_col[ind_is_num] = to_datetime({"year" : year, "month" : month, "day" : day})
                        date_col[date_col.isnull()] = pd.NaT
                        return date_col.convert_dtypes()

                if not any_valid:
                    # Convert data type to object to avoid changing dtype warning and then convert to TimeStamp
                    date_col = date_col.astype('O')
                    # This may be Epoch time
                    try:
                        date_col[ind_is_num] = to_datetime(dts, unit='s')
                    except pd._libs.tslibs.np_datetime.OutOfBoundsDatetime:
                        date_col[ind_is_num] = to_datetime(dts, unit='ms')
                    except:
                        raise

                    date_col = date_col.convert_dtypes()

                    if (date_col.dt.year > this_year).any() or (date_col.dt.year < 1971).any():
                        if not (date_col.dt.year > this_year).any() and (m:=(date_col.dt.year < 1971)).mean() < 0.01:
                            date_col[m] = pd.NaT
                        else:
                            raise ValueError("Date is outside acceptable range (1971 to this year)")
                    
                    if date_col.dt.year.max() < 1980:
                        raise ValueError("All dates are before 1980. This is unlikely to be a date column.")

                    return date_col
                    
            elif date_col.dtype in ["O", "object", "string", "string[python]"]:
                new_col = date_col.convert_dtypes()
                if new_col.dtype in ["string", "string[python]"] or \
                    str in new_col.apply(type).unique():
                    p = re.compile(r"\d{1,2}[/-]\d{1,2}[/-]\d{2,4}")
                    p2 = re.compile(r"\d{4}[/-]\d{1,2}[/-]\d{1,2}")
                    p3 = re.compile(r"\d{2}-[A-Z][a-z][a-z]-\d{2,4}", re.IGNORECASE)
                    p_not_match = re.compile(r"\d{1,2}[:\.]?\d\d[:\.]?\d?\d?")
                    num_match = 0
                    num_not_match = 0
                    k = 0
                    num_check = min(5, len(new_col))
                    for m in range(len(new_col)):
                        if isinstance(new_col[m], (pd.Timestamp, dt.datetime)):
                            num_match+=1
                            k+=1
                        elif pd.notnull(new_col[m]) and len(new_col[m].strip())!=0:
                            if p.search(new_col[m]) or p2.search(new_col[m]) or p3.search(new_col[m]):
                                num_match+=1
                            elif p_not_match.match(new_col[m])==None:
                                pass
                            else:
                                num_not_match+=1
                            k+=1

                        if k==num_check:
                            break

                    if num_match<num_check-1:
                        raise ValueError("Column is not a date column")
                    try:
                        return to_datetime(new_col)
                    except:
                        def to_dt(x):
                            try:
                                return to_datetime(x)
                            except:
                                return pd.NaT
                        new_col = new_col.apply(to_dt)
                        if new_col.isnull().sum()/len(new_col) > 0.5:
                            raise NotImplementedError()
                        else:
                            return new_col
                else:
                    raise NotImplementedError()
            else:
                raise NotImplementedError()

    return date_col


def merge_date_and_time(date_col, time_col, empty_time):
    # If date even has a time, this ignores it.
    # We assume that the time in time column is more likely to be local time
    # Often returned date is in UTC but local time is preferred for those who want to do day vs. night analysis

    # We could use:
    # return pd.to_datetime(date_col.dt.date.astype(str) + " " + time_col.astype(str), errors="coerce")
    # but not for now to catch unexpected values
    empty_time = empty_time.lower()
    def combine(d, t):
        if pd.isnull(d):
            return pd.NaT
        elif pd.isnull(t):
            if empty_time=="nat":
                return pd.NaT
            elif empty_time == "ignore":
                return d
            else:
                raise ValueError("empty_time must be 'NaT' or 'ignore'")
        else:
            return d.replace(hour=t.hour, minute=t.minute, second=t.second)

    return pd.Series([combine(d, t) for d,t in zip(date_col, time_col)])

def validate_date(df, match_cols_test):
    score = None
    match_cols = []
    for col_name in match_cols_test:
        try:
            # Fails if date column is not valid. Otherwise, returns a 
            # numerical value that is higher for more complete datetimes (i.e. date-only lower than date and time)
            date_col = parse_date_to_datetime(df[col_name])

            dts = date_col[date_col.notnull()]

            if len(dts) == 0:
                continue

            one_date = dts.iloc[0]
            max_val = 6
            new_score = None
            same_min = (dts.dt.minute == one_date.minute).all()
            if new_score is None and not same_min: 
                new_score = max_val-1
            same_hour = (dts.dt.hour == one_date.hour).all()
            if new_score is None and not same_hour: 
                new_score = max_val-2
            same_day = (dts.dt.day == one_date.day).all()
            if new_score is None and not same_day: 
                new_score = max_val-3
            same_month = (dts.dt.month == one_date.month).all()
            if new_score is None:
                if not same_month: 
                    new_score = max_val-4
                else:
                    new_score = max_val-5

            if score == new_score:
                match_cols.append(col_name)
            elif new_score != None and (score == None or new_score > score):
                # Higher scoring item found. This now takes priority
                score = new_score
                match_cols = [col_name]
        except Exception as e:
            pass

    return match_cols

    
def validate_time(df, match_cols_test, date_col=None):
    score = None
    match_cols = []
    max_len = 100000
    if len(df) > max_len:
        df = df.head(max_len)  # No need to validate everything
        date_col = date_col.head(max_len) if date_col is not None else date_col
    for col_name in match_cols_test:
        try:
            time_col = df[col_name]
            not_a_time_col_msg = "This column contains date information, and therefore, is not a time column"
            date_has_time_msg = "The date column has a time in it. There is no need for a time column."
            try:
                # Try to convert to datetime. If successful, see if there is a date value that varies.
                # If so, this is not a time column
                test_date_col = parse_date_to_datetime(time_col)
                num_unique = len(test_date_col.dt.date.unique())
                # If there are more than 2 dates or 2 dates are only 1 day apart (could be due to time zone change),
                # then the supposed time contains date information
                if num_unique > 2 or (num_unique==2 and  abs(test_date_col[0]-test_date_col[1]) > pd.Timedelta(days=1)):
                    raise ValueError(not_a_time_col_msg)
            except ValueError as e:
                if len(e.args)>0 and e.args[0] in [not_a_time_col_msg, date_has_time_msg]:
                    continue
            except Exception:
                pass

            new_time_col = parse_time(time_col)

            try: 
                if date_col is not None:
                    date_times = date_col.dt.time
                    # Find times from the date column that differ from the time column
                    date_times = date_times[
                        date_col.dt.time.apply(lambda x: x.replace(second=0) if pd.notnull(x) else pd.NaT) != 
                        new_time_col.apply(lambda x: x.replace(second=0) if pd.notnull(x) else pd.NaT)]
                    counts = date_times.value_counts()
                    if len(counts)==0:
                        # All times are the same between date and time columns
                        raise ValueError(date_has_time_msg)
                    most_common_time = date_times.mode()

                    # If most common time is all zeros, assume no time in date
                    if most_common_time[0]!=dt.time(hour=0, minute=0, second=0):
                        # If the date has no time, it will have zeros, standard time offsets, or DST offsets
                        if len(counts)>3 or any([x.minute!=0 or x.second!=0 for x in counts.index]):
                            # Check the percentage of times in the date column whose minute is 0. It should be relatively rare
                            if sum([counts[x] for x in counts.index if x.minute!=0]) / counts.sum() > 0.4:
                                raise ValueError(date_has_time_msg)
                        else:
                            # 00:00 must be one of the numbers if 3 unique values
                            if len(counts)==3 and all([x!=dt.time(hour=0, minute=0, second=0) for x in counts.index]):
                                raise ValueError(date_has_time_msg)
                            
                            non_zero = [x for x in counts.index if x.hour!=0]
                            if len(non_zero)>1:
                                # Values should be off by one hour
                                d1 = dt.timedelta(hours=non_zero[0].hour)
                                d2 = dt.timedelta(hours=non_zero[1].hour)
                                if d2-d1 != dt.timedelta(hours=1) and d1-d2 != dt.timedelta(hours=1):
                                    raise ValueError(date_has_time_msg)
            except ValueError as e:
                if len(e.args)>0 and e.args[0] in [not_a_time_col_msg, date_has_time_msg]:
                    continue
            except Exception:
                pass

            new_time_col = new_time_col[new_time_col.notnull()]
            
            new_score = None
            if len(new_time_col) == 0:
                continue

            hours = pd.Series([t.hour if pd.notnull(t) else np.nan for t in new_time_col])
            mins = pd.Series([t.minute if pd.notnull(t) else np.nan for t in new_time_col])
            # secs = time_sec - hours*3600 - mins*60
            max_val = 3
            # same_sec = (secs == secs.iloc[0]).all()
            # if not same_sec: 
            #     return max_val
            same_min = (mins == mins.iloc[0]).all()
            if not same_min: 
                new_score = max_val-1
            same_hour = (hours == hours.iloc[0]).all()
            if new_score is None:
                if not same_hour: 
                    new_score = max_val-2
                else:
                    new_score = max_val-3
                
            if score == new_score:
                match_cols.append(col_name)
            elif new_score != None and (score == None or new_score > score):
                # Higher scoring item found. This now takes priority
                score = new_score
                match_cols = [col_name]

        except Exception as e:
            pass

    return match_cols


def parse_time(time_col):
    # Returns time in seconds since 00:00
    time_col = time_col.copy()
    if pd.api.types.is_numeric_dtype(time_col):
        # Expected to be time as integer in 24-hr HHMM format
        # Check that this is true
        time_col[(time_col==9999) | (time_col==999)] = pd.NA  # 9999 is used as error code
        hour = np.floor(time_col/100)
        min = time_col - np.floor(time_col/100)*100
        if hour.max() >= 24:
            invalid = hour>=24
            if invalid.mean() < 0.01:
                # These are likely recording errors. Replace with NaN
                hour.loc[invalid] = np.nan
            else:
                raise NotImplementedError()
        if min.max() > 59:
            invalid = min>=60
            if invalid.mean() < 0.01:
                # These are likely recording errors. Replace with NaN
                min.loc[invalid] = np.nan
            else:
                raise NotImplementedError()

        return pd.Series([dt.time(hour=int(x),minute=int(y)) if (pd.notnull(x) and pd.notnull(y)) else pd.NaT for x,y in zip(hour,min)])
    elif time_col.dtype in ['O','string[python]']:
        new_col = time_col.convert_dtypes()
        if new_col.dtype in ["string",'string[python]'] or \
            time_col.apply(lambda x: isinstance(x,str) or isinstance(x,int) or pd.isnull(x) or isinstance(x,dt.time)).all():
            # Cleanup split AM or PM, which causes a warning from pandas
            new_col = new_col.apply(lambda x: x.replace("P M","PM").replace("A M","AM") if isinstance(x,str) else x)
            new_col = new_col.apply(lambda x: x.strftime('%H:%M') if isinstance(x,dt.time) else x)
            try:
                new_col = to_datetime(new_col)
                return new_col.dt.time
            except:
                pass

            p_date = [re.compile(r"\d{1,2}[/-]\d{1,2}[/-]\d{2,4}"),
                      re.compile(r"^\d{1,2}-(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)$"),
                      re.compile(r"^(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)-\d{1,2}$")]
            def convert_timestr_to_sec(x):
                if pd.isnull(x):
                    return x
                
                if isinstance(x,int):
                    hour = np.floor(x/100)
                    min = x - np.floor(x/100)*100
                    if hour >= 24:
                        raise NotImplementedError()
                    if min > 59:
                        raise NotImplementedError()
                    return dt.time(hour=int(hour),minute=int(min))

                x = x.replace(" ","")

                if (x.upper().endswith('AM') or x.upper().endswith('PM')) and \
                    x[:-2].isdigit():
                    # Time does not have colon. Try to insert one
                    x = x[:-4] + ":" + x[-4:]
                    
                time_list = x.split(":")
                if len(time_list)==1 and len(x.split("."))>1:
                    time_list = x.split(".")
                    if len(time_list)!=3:
                        raise NotImplementedError()

                if len(time_list)==1 and len(x)==5 and x[2]==';':
                    # Accidentally used ; instead of :
                    time_list = x.split(";")
                if len(time_list)==1:
                    if x.strip() in ["","-"]:
                        return pd.NaT
                    elif len(x) == 0 or len(x) > 4 or not x.isdigit():
                        if x in ["#NAME?",'#VALUE!', 'TIME','NULL'] or re.search(r'^C\d+',x) or \
                            any([y.search(x) for y in p_date]):  # Date accidently entered in time column
                            # C2 values were observed in 1 dataset
                            return pd.NaT
                        else:
                            raise ValueError("Expected HHMM format")

                    min = float(x[-2:])
                    if min > 59:
                        return pd.NaT
                    
                    if len(x) > 2:
                        hour = float(x[:-2])
                    else:
                        hour = 0

                    if hour > 23:
                        return pd.NaT

                    return dt.time(hour=int(hour),minute=int(min))

                if "T" in time_list[0]:
                    t_loc = [k for k,x in enumerate(time_list[0]) if x=="T"]
                    time_list[0] = time_list[0][t_loc[0]+1:]

                hours_add = 0
                if "AM" in time_list[-1].upper():
                    time_list[-1] = time_list[-1].upper().replace("AM", "")
                    if time_list[0].strip() == "12":
                        time_list[0] = "0"
                elif "PM" in time_list[-1].upper():
                    hours_add = 12
                    time_list[-1] = time_list[-1].upper().replace("PM", "")
                    if time_list[0].strip() == "12":
                        time_list[0] = "0"
                    elif int(time_list[0].strip()) > 12:
                        # Typo where 24 hour time was provided but also PM indicated
                        hours_add = 0

                try:
                    t = dt.time(hour=int(time_list[0])+hours_add,minute=int(time_list[1]))
                except:
                    return pd.NaT

                if len(time_list) > 2:
                    try:
                        t = t.replace(second=int(time_list[2]))
                    except:
                        # One case found where seconds are XX
                        pass

                return t
            return new_col.apply(convert_timestr_to_sec)
        else:
            raise NotImplementedError()
    else:
        raise NotImplementedError()
    

def to_datetime(dates, ignore_errors=False, *args, **kwargs):
    coerce = 'errors' in kwargs and kwargs['errors']=='coerce'
    def to_datetime_local(x, *args, **kwargs):
        try:
            return pd.to_datetime(x, *args, **kwargs)
        except:
            return pd.NaT if coerce else x  # Coerce input to pd.to_datetime should return NaT if cannot convert to datetime

    if isinstance(dates, pd.DataFrame) and \
        (dates.isnull().any().any() or dates.apply(lambda x: isinstance(x,str)).any()):
        # This should have columns year, month, and day
        # Special code to handle nans in month or day which will be set to Periods instead of datetimes
        # or to handle cases where the day is an ordinal
        # This is year/month/date with NaNs in some of the values
        year_col = [x for x in dates.columns if x.lower()=='year'][0]
        month_col = [x for x in dates.columns if x.lower()=='month'][0]
        day_col = [x for x in dates.columns if x.lower()=='day'][0]
        def to_datetime_local(x):
            if pd.isnull(x[day_col]):
                if pd.isnull(x[month_col]):
                    if pd.isnull(x[year_col]):
                        return pd.NaT
                    else:
                        return pd.Period(freq='Y', year=x[year_col])
                else:
                    return pd.Period(freq='M', year=x[year_col], month=x[month_col])
            elif pd.isnull(x[month_col]):
                return ValueError(f"Month is null but day is not. Unable to parse {x}")
            elif pd.isnull(x[year_col]):
                return ValueError(f"Year is null but month and day are not. Unable to parse {x}")
            
            return pd.to_datetime(x.to_frame().T, *args, **kwargs).iloc[0]
        
        # Replace any ordinal days
        try:
            dates[day_col] = dates[day_col].apply(lambda x: x.strip()[:-2] if  isinstance(x,str) and re.search(r'\d+(th|st|nd|rd)',x) else x)
            return dates.apply(to_datetime_local, axis=1)
        except:
            # Adding this to put breakpoint here to catch errors. This try/except should be removed in the future
            raise

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="Could not infer format")
        try:
            return pd.to_datetime(dates, *args, **kwargs)
        except UnicodeEncodeError as e:
            if (ignore_errors or 'errors' in kwargs and kwargs['errors']=='coerce') and isinstance(dates, pd.Series):
                return dates.apply(to_datetime_local)
            else:
                return dates
        except (pd._libs.tslibs.parsing.DateParseError, dateutil.parser._parser.ParserError) as e:
            if 'out of range' in str(e) and \
                ((m1:=re.search(r'year ((19|20)\d{6}) is out of range: \1\.0.* position 0', str(e))) or \
                 re.search(r'year (\d{3,4}(19|20)\d{2}) is out of range: 0?\1\.0.* position 0', str(e))):
                if isinstance(dates, pd.Series):
                    return dates.apply(to_datetime)
                elif m1:
                    return pd.to_datetime(dates[:-2])
                else:
                    dates = dates[:-2]
                    year = dates[-4:]
                    mmdd = dates[:-4] if len(dates)==8 else '0'+dates[:-4]
                    return pd.to_datetime(year+mmdd)
            elif re.search(r'\d{2}/\d{2}/\d{4} \d{4} hours', str(e)) and isinstance(dates, pd.Series):
                return pd.to_datetime(dates, format='%m/%d/%Y %H%M hours')
            elif ignore_errors and 'out of range' in str(e) and not isinstance(dates, pd.Series):
                return dates
            elif ignore_errors and isinstance(dates,pd.Series) and \
                  (re.search(r'(Unknown string format|unable to parse): \d{4}\-__', str(e)) or \
                   re.search(r'(Unknown string format|unable to parse): \d{4}\-\d{2}\-__', str(e))):
                return dates.apply(lambda x: to_datetime(x, ignore_errors, *args, **kwargs))
            elif ignore_errors and isinstance(dates,str) and \
                ((m:=re.search(r'^(?P<year>\d{4})\-(?P<month>[\d\_]{2})\-\_\_', dates)) or \
                 (m:=re.search(r'^(?P<year>\d{4})\-(?P<month>\_\_)\-\d{2}', dates))):
                if m.groupdict()['month'].isdigit():
                    return pd.Period(freq='M', year=int(m.groupdict()['year']), month=int(m.groupdict()['month']))
                else:
                    return pd.Period(freq='Y', year=int(m.groupdict()['year']))
            elif ignore_errors and isinstance(dates, str) and re.search(r'Unknown(\sdatetime)? string format.+: \d+[-/]\d+[-/]\d+,?\s?\d+[-/]\d+[-/]\d+', str(e)):
                # Comma separated list of dates. Just return value
                return dates
            else:
                raise
        except ValueError as e:
            if ignore_errors and "Given date string" in str(e) and "not likely a datetime" in str(e) and \
                (len(args)>0 or 'errors' not in kwargs or kwargs['errors']!='coerce'):
                new_kwargs = kwargs.copy()
                new_kwargs['errors'] = 'coerce'
                dts = pd.to_datetime(dates, *args, **new_kwargs)
                raise NotImplementedError()
            elif ignore_errors and isinstance(dates,pd.Series) and \
                    (re.search(r'time data \"\d{4}\-(\d\d|\_\_)\-\_\_\"',  str(e)) or \
                     re.search(r'time data \"\d{4}\-\_\_\-\d\d\"',  str(e))):
                def to_datetime_local(x):
                    if isinstance(x, str) and ((m:=re.search(r'^(?P<year>\d{4})\-(?P<month>[\d\_]{2})\-\_\_', x)) or \
                        (m:=re.search(r'^(?P<year>\d{4})\-(?P<month>[\_]{2})\-\d{2}', x))):
                        if m.groupdict()['month'].isdigit():
                            return pd.Period(freq='M', year=int(m.groupdict()['year']), month=int(m.groupdict()['month']))
                        else:
                            return pd.Period(freq='Y', year=int(m.groupdict()['year']))
                    else:
                        return pd.to_datetime(x, *args, **kwargs)
                    
                return dates.apply(to_datetime_local)
            elif 'doesn\'t match format "%Y-%m-%dT%H:%M:%S.%f%z"' in str(e):
                return pd.to_datetime(dates.apply(lambda x: x[:-1] if isinstance(x,str) and x[-1]=='Z' else x), format='mixed')
            elif ignore_errors and 'is out of range' in str(e) and is_str_number(dates):
                if (dec:=dates.find('.'))>0 and int(dates[dec+1:])==0:
                    dates = int(dates[:dec])
                    year_first = math.floor(dates/10000)
                    year_last = dates % 10000

                    is_year_first = 1980 <= year_first <= datetime.datetime.now().year
                    is_year_last =  1980 <= year_last <= datetime.datetime.now().year
                    if is_year_first + is_year_last != 1:
                        raise e
                    elif is_year_first:
                        year = year_first
                        month_day = dates - year*10000
                        month = math.floor(month_day/100)
                        day = month_day - month*100
                    else:
                        year = year_last
                        month_day = year_first
                        month = math.floor(month_day/100)
                        day = month_day - month*100
                    if month>12 or day>31:
                        raise e
                    return pd.to_datetime(datetime.datetime(year=year, month=month, day=day))
                else:
                    raise e
            elif ignore_errors and re.search(r'Unknown string format: \d+[-/]\d+[-/]\d+,?\s?\d+[-/]\d+[-/]\d+', str(e)):
                # Comma separated list of dates. Just return value
                return dates
            elif isinstance(dates, pd.DataFrame) and dates.shape[1]==3 and (str(e)=='cannot convert NA to integer' or \
                    (re.search(r'Unable to parse string \"\d+(th|st|nd|rd)\"', str(e)) and \
                    dates.iloc[:,2][dates.iloc[:,2].notnull()].apply(lambda x: isinstance(x, numbers.Number) or is_str_number(x) or \
                    re.search(r'\d+(th|st|nd|rd)',x) is not None).all())):
                raise NotImplementedError("This should no longer be reachable. Remove in a future release")
            else:
                return pd.to_datetime(dates, *args, format='mixed', **kwargs)
        except Exception as e:
            raise


from ._version import __version__
from .data import Source
from . import defs
from .import datasets
from .defs import TableType
from .defs import columns as Column
from .deprecated.datasetsCompat import datasets_query


import pandas as pd
import numpy as np
import os
import re
from packaging import version
import pytest

if __name__ == "__main__":
	import sys
	sys.path.append('../openpolicedata')
import openpolicedata as opd

import pathlib
import sys
sys.path.append(pathlib.Path(__file__).parent.resolve())
from test_utils import get_datasets


local_csv_file = os.path.join("..",'opd-data','opd_source_table.csv')

@pytest.mark.parametrize("file", [(), (local_csv_file,)])
def test_reload(file):
    orig = opd.datasets.datasets
    opd.datasets.datasets = None
    assert opd.datasets.datasets is None
    try:
        if len(file)>0:
            if not os.path.exists(os.path.dirname(local_csv_file)):
                return
            assert os.path.exists(*file)
        opd.datasets.reload(*file)
        assert isinstance(opd.datasets.datasets, pd.DataFrame)
    except:
        raise
    finally:
        opd.datasets.datasets = orig


def test_duplicates(all_datasets):
    assert not all_datasets.duplicated(subset=['State', 'SourceName', 'Agency', 'TableType','Year','source_url','dataset_id']).any()


def test_check_columns(datasets):
    columns = {
        'State' : pd.StringDtype(),
        'SourceName' : pd.StringDtype(),
        'Agency': pd.StringDtype(),
        'TableType': pd.StringDtype(),
        'Year': np.dtype("O"),
        'Description': pd.StringDtype(),
        'DataType': pd.StringDtype(),
        'URL': pd.StringDtype(),
        'date_field': pd.StringDtype(),
        'dataset_id': pd.StringDtype(),
        'agency_field': pd.StringDtype()
    }

    for key in columns.keys():
        assert key in datasets


def test_table_for_nulls(datasets):
    can_have_nulls = ["Description", "date_field", "dataset_id", "agency_field", "Year","readme","min_version",
                        "AgencyFull","source_url","coverage_start","coverage_end",'query']

    for col in datasets.columns:
        if not col in can_have_nulls:
            assert pd.isnull(datasets[col]).sum() == 0


def test_check_state_names(datasets):
    all_states = [
        'Alabama', 'Alaska', 'American Samoa', 'Arizona', 'Arkansas', 'California', 'Colorado', 'Connecticut', 'Delaware', 'District of Columbia',
        'Florida', 'Georgia', 'Guam', 'Hawaii', 'Idaho', 'Illinois', 'Indiana', 'Iowa', 'Kansas', 'Kentucky', 'Louisiana', 'Maine',
        'Maryland', 'Massachusetts', 'Michigan', 'Minnesota', 'Mississippi', 'Missouri', 'Montana', 'Nebraska', 'Nevada', 'New Hampshire',
        'New Jersey', 'New Mexico', 'New York', 'North Carolina', 'North Dakota', 'Northern Mariana Islands', 'Ohio', 'Oklahoma', 'Oregon',
        'Pennsylvania', 'Puerto Rico', 'Rhode Island', 'South Carolina', 'South Dakota', 'Tennessee', 'Texas', 'Utah', 'Vermont', 'Virgin Islands',
        'Virginia', 'Washington', 'West Virginia', 'Wisconsin', 'Wyoming'
    ]

    assert len([x for x in datasets["State"] if x not in all_states]) == 0

def test_agency_names(datasets):
    # Agency names should either match source name or be MULTI
    rem = datasets["Agency"][datasets["Agency"] != datasets["SourceName"]]
    assert ((rem == opd.defs.MULTI) | (rem == opd.defs.NA)).all()

def test_year(datasets):
    # year should either be an int or MULTI or "None"
    rem = datasets["Year"][[type(x)!=int for x in datasets["Year"]]]
    assert ((rem == opd.defs.MULTI) | (rem == opd.defs.NA)).all()

@pytest.mark.parametrize('data_type', [opd.defs.DataType.SOCRATA, opd.defs.DataType.CARTO, opd.defs.DataType.CKAN])
def test_dataset_id(datasets, data_type):
    rem = datasets["dataset_id"][datasets["DataType"] == data_type]
    assert pd.isnull(rem).sum() == 0

def test_years_multi(datasets):
    # Multi-year datasets should typically have a value in date_field
    datasets = datasets[datasets["Year"] == opd.defs.MULTI]
    df_null = datasets[pd.isnull(datasets["date_field"])]
    
    # This can only be allowed for certain Excel cases
    assert (df_null["DataType"] == opd.defs.DataType.EXCEL.value).all()

def test_agencies_multi(datasets):
    rem = datasets["agency_field"][datasets["Agency"] == opd.defs.MULTI]
    assert pd.isnull(rem).sum() == 0

def test_arcgis_urls(datasets):
    urls = datasets["URL"]
    p = re.compile(r"(MapServer|FeatureServer)/\d+")
    for i,url in enumerate(urls):
        if datasets.iloc[i]["DataType"] == opd.defs.DataType.ArcGIS.value:
            result = p.search(url)
            assert result != None
            assert len(url) == result.span()[1]

def test_source_list_by_state(datasets, use_changed_rows):
    state = "Virginia"
    df = opd.datasets.query(state=state)
    df_truth = datasets[datasets["State"]==state]
    assert use_changed_rows or len(df)>0
    pd.testing.assert_frame_equal(df_truth, df)

def test_source_list_by_source_name(datasets, use_changed_rows):
    source_name = "Fairfax County"
    df = opd.datasets.query(source_name=source_name)
    df_truth = datasets[datasets["SourceName"]==source_name]
    assert use_changed_rows or len(df)>0
    pd.testing.assert_frame_equal(df_truth, df)

def test_source_list_by_agency(datasets, use_changed_rows):
    agency = "Fairfax County"
    df = opd.datasets.query(agency=agency)
    df_truth = datasets[datasets["Agency"]==agency]
    assert use_changed_rows or len(df)>0
    pd.testing.assert_frame_equal(df_truth, df)

def test_source_list_by_table_type(datasets, use_changed_rows):
    table_type = opd.defs.TableType.TRAFFIC
    df = opd.datasets.query(table_type=table_type)
    df_truth = datasets[datasets["TableType"]==table_type.value]
    assert use_changed_rows or len(df)>0
    pd.testing.assert_frame_equal(df_truth, df)

def test_source_list_by_table_type_value(datasets, use_changed_rows):
    table_type = opd.defs.TableType.TRAFFIC.value
    df = opd.datasets.query(table_type=table_type)
    df_truth = datasets[datasets["TableType"]==table_type]
    assert use_changed_rows or len(df)>0
    pd.testing.assert_frame_equal(df_truth, df)

def test_source_list_by_multi(datasets, use_changed_rows):
    state = "Virginia"
    source_name = "Fairfax County"
    table_type = opd.defs.TableType.TRAFFIC_CITATIONS.value
    df = opd.datasets.query(state=state, table_type=table_type, source_name=source_name)
    df_truth = datasets[datasets["TableType"]==table_type]
    df_truth = df_truth[df_truth["State"]==state]
    df_truth = df_truth[df_truth["SourceName"]==source_name]
    assert use_changed_rows or len(df)>0
    pd.testing.assert_frame_equal(df_truth, df)

def test_table_types(datasets):
    for t in datasets["TableType"]:
        # Try to convert to an enum
        opd.defs.TableType(t)

def test_data_types(datasets):
    for t in datasets["DataType"].unique():
        # Try to convert to an enum
        opd.defs.DataType(t)

def test_min_versions(datasets):
    for ver in datasets["min_version"][datasets["min_version"].notnull()]:
        if not (ver == "-1" or type(version.parse(ver)) == version.Version):
            raise ValueError(f"{ver} is an invalid value for min_version")
        
def test_summary_functions():
    opd.datasets.num_unique()
    opd.datasets.num_sources()
    opd.datasets.num_sources(full_states_only=True)
    opd.datasets.summary_by_state()
    opd.datasets.summary_by_state(by="YEAR")
    opd.datasets.summary_by_state(by="TABLE")
    opd.datasets.summary_by_table_type()
    opd.datasets.summary_by_table_type(by_year=True)

    opd.datasets.datasets.loc[0,"TableType"] = "TEST"
    with pytest.warns(UserWarning):
        opd.datasets.summary_by_table_type()

def test_get_table_types(all_datasets):   # Passing in all_datasets to ensure that local changes are used if csvfile is set 
    opd.datasets.get_table_types()
    stops_tables = opd.datasets.get_table_types(contains="STOPS")
    exp_stops_tables = ["PEDESTRIAN STOPS","STOPS","TRAFFIC STOPS",
                        "TRAFFIC STOPS - INCIDENTS", 'TRAFFIC STOPS - SUBJECTS',
                        "STOPS - INCIDENTS", 'STOPS - SUBJECTS']
    assert len(stops_tables) == len(exp_stops_tables)
    assert all([x in exp_stops_tables for x in stops_tables])

def test_combined_num_datasets(all_datasets):
    for k in range(len(all_datasets)):
        if pd.notnull(all_datasets.iloc[k]['dataset_id']) and ';' in all_datasets.iloc[k]['dataset_id']:
            ds = all_datasets.iloc[k]['dataset_id']
            if '|' in ds:  # Before | are sheet names which we don't need
                ds = ds.split('|')[1]
            ds = [x.strip() for x in ds.split(';')]
            if all_datasets.iloc[k]['SourceName']=='Wallkill' and all_datasets.iloc[k]['Year']==2016:
                assert len(ds)==3
            else:
                # Data quality checks (not set in stone. may need to adjust for new cases)
                # Biannually, Quarterly, or monthly
                assert len(ds) in [2, 4,12]
            # No repeats
            assert len(ds) == len(set(ds))


if __name__ == "__main__":
    csvfile = None
    csvfile = os.path.join('..', 'opd-data', "opd_source_table.csv")
    test_dataset_id(csvfile,None,None,None,None)
    # test_get_table_types(csvfile,None,None,None,None)
    # test_table_for_nulls(csvfile,None,None,None,None)
    # test_years_multi(csvfile,None,None,None,None)
    # test_table_types(csvfile,None,None,None,None)
    # test_agencies_multi(csvfile,None,None,None,None)
    # test_reload(csvfile,None,None,None,None, (csv_file,))


from enum import Enum
from numbers import Number
import numpy as np
import pandas as pd
import re

from .utils import camel_case_split
from . import defs

class DataMapping:
    def __init__(self, orig_column_name=None, new_column_name = None, data_maps=None, orig_column=None):
        self.orig_column_name = orig_column_name
        self.new_column_name = new_column_name
        self.data_maps = data_maps
        self.orig_value_counts = orig_column.value_counts().head() if orig_column is not None else None

    def __repr__(self, ) -> str:
        return ',\n'.join("%s: %s" % item for item in vars(self).items())
    
    def __eq__(self, other): 
        if not isinstance(other, DataMapping):
            return False
        
        tf_data_maps = self.data_maps == other.data_maps
        if not tf_data_maps:
            if (self.data_maps is None and other.data_maps is not None) or \
                (self.data_maps is not None and other.data_maps is None):
                return False
            if len(self.data_maps)==len(other.data_maps):
                for k in self.data_maps.keys():
                    if k not in other.data_maps and \
                        not (pd.isna(k) and any([pd.isna(x) for x in other.data_maps.keys()])):
                        return False
                    if pd.isnull(k):
                        kother = [x for x in other.data_maps.keys() if pd.isnull(x)][0]
                    else:
                        kother = k
                    if self.data_maps[k] != other.data_maps[kother]:
                        return False
                
                tf_data_maps = True
            elif set(self.data_maps.keys()).symmetric_difference(set(other.data_maps.keys())) == {np.nan}:
                # Adding nan to data maps later
                if (np.nan in other.data_maps and other.data_maps[np.nan]=="UNSPECIFIED") or \
                    (np.nan in self.data_maps and self.data_maps[np.nan]=="UNSPECIFIED"):
                    tf_data_maps = True


        tf_vals = self.orig_value_counts is None and other.orig_value_counts is None
        if not tf_vals and self.orig_value_counts is not None and other.orig_value_counts is not None:
            tf_vals = self.orig_value_counts.equals(other.orig_value_counts)
            if not tf_vals and other.orig_value_counts.index.dtype=="int64" and all([x.isdigit() for x in self.orig_value_counts.index if isinstance(x,str)]):
                # Found a case where indices were numeric strings but read back in as numbers
                other.orig_value_counts.index = [str(x) for x in other.orig_value_counts.index]
                tf_vals = self.orig_value_counts.equals(other.orig_value_counts)
            if not tf_vals and len(self.orig_value_counts)==len(other.orig_value_counts) and \
                all(self.orig_value_counts.values == other.orig_value_counts.values):
                tf_vals = True
                for x,y in zip(self.orig_value_counts.index, other.orig_value_counts.index):
                    # Difference of trailing 0 is OK
                    if x!=y:
                        if isinstance(x,str) and isinstance(y,str) and (x=="0"+y or y=="0"+x):
                            continue
                        elif isinstance(x,str) and x.isdigit() and isinstance(y,Number) and float(x)==y:
                            continue
                        elif isinstance(y,str) and y.isdigit() and isinstance(x,Number) and float(y)==x:
                            continue
                        elif isinstance(y,str) and isinstance(x, pd._libs.tslibs.timestamps.Timestamp) and \
                            (y==x.strftime("%Y-%m-%d %H:%M:%S") or (y==x.strftime("%Y-%m-%d") and x.hour==0 and x.minute==0)):
                            continue
                        elif isinstance(x,str) and isinstance(y, pd._libs.tslibs.timestamps.Timestamp) and \
                            (x==y.strftime("%Y-%m-%d %H:%M:%S") or (x==y.strftime("%Y-%m-%d") and y.hour==0 and y.minute==0)):
                            continue
                        elif isinstance(y,str):
                            try:
                                float(y)==x
                                continue
                            except:
                                pass
                        tf_vals = False

        new_name_equal = self.new_column_name == other.new_column_name or self.new_column_name=="RAW_"+other.new_column_name
                        
        tf = self.orig_column_name == other.orig_column_name and \
            new_name_equal and \
            tf_data_maps and tf_vals

        return tf


def check_column(col_name, col_types):
    if isinstance(col_types, str):
        col_types = [col_types]

    for col_type in col_types:
        col_type = col_type.lower()

        # Check for unambiguous column names
        if col_name.lower()==col_type:
            return True

        # Check if column name is a descriptive term + race
        # sus has been used as shorthand for suspect
        # ofc has been used as shorthand for officer
        desc_terms = ["citizen","subject","suspect", "sus", "civilian", "complainant", "person", "cit",
                      "offender", "officer", "ofc", "deputy", "off", 'ctzn']

        if any([col_name.lower()==x+col_type or col_name.lower()==col_type+x for x in desc_terms]):
            return True
        
        words = set(re.split(r"[^A-Za-z]+", col_name.lower()))
        if any([{x,col_type}.issubset(words) for x in desc_terms]):
            return True

        words = set(camel_case_split(col_name))
        if any([{x,col_type}.issubset(words) for x in desc_terms]):
            return True
    
    return False
    

class _case:
    def __init__(self, src, table_type, old_name, new_name, year=None):
        self.src = src
        self.table_type = table_type
        self.year = year
        self.old_name = old_name if type(old_name)==list else [old_name]
        self.new_name = new_name if type(new_name)==list else [new_name]

    def __repr__(self, ) -> str:
        return ',\n'.join("%s: %s" % item for item in vars(self).items())

    def equals(self, src, table_type, year):
        tf = src==self.src and table_type==self.table_type
        if self.year!=None:
            if isinstance(self.year, Number):
                tf = tf and (year==self.year)
            else:
                tf = tf and (year in self.year)

        return tf
    
    def findcols(self, columns):
        for k in range(len(self.old_name)):
            if isinstance(self.old_name[k],list):
                if any(x not in columns for x in self.old_name[k]):
                    return False
            elif self.old_name[k] not in columns:
                return False
        return True


class MultType(Enum):
    SINGLE = 0
    DICT = 1
    DEMO_COL = 2
    COUNTS = 3
    DELIMITED = 4
    WITH_NAME = 5
    WITH_COUNTS = 6

class _MultData:
    type = MultType.SINGLE
    delim_race = None
    delim_age = None
    delim_gender = None
    delim_eth = None
    item_race = None
    item_age = None
    item_gender = None
    item_eth = None

    def __repr__(self, ) -> str:
        return '\n'.join("%s: %s" % (name, getattr(self,name)) for name in dir(self) if not name.startswith("__"))
